!=========================================================================
! Copyright (C) GemTalk Systems 1986-2013.  All Rights Reserved.
!
! $Id: reposit.gs,v 1.15 2008-03-10 17:41:37 bille Exp $
!
! Superclass Hierarchy:
!   Repository, Array, SequenceableCollection, Collection, Object.
!
!=========================================================================

removeallmethods Repository
removeallclassmethods Repository

category: 'For Documentation Installation only'
classmethod: Repository
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A Repository is an object that represents a virtual storage into which users
 can place their data.  Each Repository includes an Array of Segments,
 which represent authorization regions for the data.  
 Repositories are described in the GemStone Programming Guide.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A Symbol; the user-supplied logical name for the Repository.'.
doc documentInstVar: #name with: txt.

txt := (GsDocText new) details:
'A dictionary that describes the logical entry points to data in this
 Repository; a convenient mechanism for remembering and accessing the objects
 in the Repository, similar to a file system directory. (Reserved for future
 use.)'.
doc documentInstVar: #dataDictionary with: txt.

txt := (GsDocText new) details:
'Backups and restoration are ordinarily performed while using the GemStone
 DataCurator login.  It is possible to use another login that also has the
 FileControl privilege.  However, for restorations, it is recommended that you
 use only the DataCurator or the SystemUser logins.  If you use another login,
 and that login disappears as a result of the restoration, you will see a
 fatal error.

 The GemStone System Administration Guide discusses backups and restoration
 in more detail.'.
doc documentCategory: #'Backup and Restore' with: txt.

txt := (GsDocText new) details:
'Methods in this category are obsolete and are provided only for compatibility
 with earlier releases of GemStone.  They will be removed in a future release.'.
doc documentCategory: #'Backward Compatibility' with: txt.

self description: doc.
%

!==========================================================================
! Below this point: only method definitions in fileout order
!==========================================================================

! ------------------- Class methods for Repository
category: 'Instance Creation'
classmethod: Repository
new

"Disallowed."

self shouldNotImplement: #new
%
category: 'Instance Creation'
classmethod: Repository
new: anInteger

"Disallowed."

self shouldNotImplement: #new:
%
! ------------------- Instance methods for Repository
category: 'Backup and Restore'
method: Repository
abortFullBackup

 <primitive: 474>

"Cancel a full backup that is in progress.  If fullBackupTo: has been used to
 start a multi-file backup, but continueFullBackupTo:MBytes: has not been
 executed to completion of the backup, you can use abortFullBackup to cancel
 the full backup and permit this session to commit and abort.

 This method requires no privileges.  If a backup is not in progress, this
 has no effect."

  self _primitiveFailed: #abortFullBackup.
  self _uncontinuableError
%
category: 'Backup and Restore'
method: Repository
abortRestore

"If a restore from backups is in progress on multi-file backup, this method
 cancels the restore, and reenables logins.  The Repository reverts to the
 state prior to starting the restore.

 If the last file of a backup has already been restored, or a restore from
 a multifile backup is not in progress, this method has no effect.

 Note that this method has no effect if a restore from backup has completed
 and restore from transaction logs is in progress.  To stop restoring
 transaction logs you must use commitRestore.

 Returns true.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.

 If a backup file read with restoreFromBackup: is truncated or corrupt, 
 it may be necessary to execute abortRestore before restoreFromBackup:
 can be used to restart the restore from a good copy of the backup file."

^ self _restoreFrom: nil opcode: 1
%
category: 'Adding'
method: Repository
add: newObject

"Disallowed."

self shouldNotImplement: #add:
%
category: 'Adding'
method: Repository
addAll: aCollection

"Disallowed."

self shouldNotImplement: #addAll:
%
category: 'Garbage Collection'
method: Repository
addGcCandidates: anArray

"Adds an Array of candidates to the global queue of GcCandidates.  When the
 transaction that this method is executed in is committed, the candidates will 
 be visible to a session that performs the markGcCandidates method."
   
(anArray isKindOf: Array)
  ifFalse: [^ anArray _errorExpectedClass: Array].

(Globals at: #GcCandidates) add: anArray.
(Globals at: #GcCandidatesCount) incrementBy: (anArray size).
%
category: 'Adding'
method: Repository
addLast: newObject

"Disallowed."

self shouldNotImplement: #addLast:
%
category: 'Transaction Logging'
method: Repository
addTransactionLog: deviceOrDirectory  replicate: replicateSpec size: aSize
 
"Add deviceOrDirectory to the configuration parameter STN_TRAN_LOG_DIRECTORIES.
 If STN_TRAN_LOG_REPLICATES is not empty, then replicateSpec must be a valid
 device or directory.  If STN_TRAN_LOG_REPLICATES is empty (transaction log
 replicates are not being used), then replicateSpec must be an empty String.
 
 The aSize argument must be a positive SmallInteger; it is added to the value
 of the STN_TRAN_LOG_SIZES configuration parameter.

 This method requires the FileControl privilege."

<primitive: 337>

deviceOrDirectory _validateClass: String .
replicateSpec ~~ nil ifTrue:[ replicateSpec _validateClass: String ] .
aSize _validateClass: SmallInteger .
aSize > 0 ifFalse:[ aSize _error: #rtErrArgOutOfRange ] .
^ self _primitiveFailed: #addTransactionLog:replicate:size: 
%
category: 'Updating'
method: Repository
at: anIndex put: aValue

"Disallowed."

self shouldNotImplement: #at:put:
%
category: 'Updating'
method: Repository
atAllPut: anObject

"Disallowed."

self shouldNotImplement: #atAllPut:
%
category: 'Garbage Collection'
method: Repository
auditWithLimit: sizeLimit

"Checks all objects in GemStone for consistency.  (Compare with Repository's
 instance method repairWithLimit:.)  A description of errors found is written to
 standard output, along with statistics about the Repository.  The
 statistics report will not include any objects smaller than the specified
 sizeLimit (number of bytes or OOPs).

 This method should be executed from topaz -l (the linked version of Topaz).

 This method aborts the current transaction.

 This method requires the GarbageCollection privilege.

 The Garbage Collector session is shut down for the duration of this method.

 A GciHardBreak during this method terminates the session.

 If this session is the only session logged, logins are disabled for the
 duration of this method, scavenging is forced to complete, and additional
 consistency checks are made during the audit.   If other users were found to be
 logged in, then scavenging is not completed, logins are not disabled, and the
 audit performs less checking of the Repository.

 If for some reason scavenging cannot be completed, this method runs as if
 other users were logged in, and the audit performs less checking of the 
 Repository. Refer to auditWithLimit:fullChecks: for a variation that will 
 return immediately with errors rather than performing the audit  with 
 fewer checks.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 This method raises the special errors #3021 or #3032.  Topaz performs 
 special processing over these results to determine whether or not the audit 
 succeeded.  The Topaz  'expectvalue true' command matches the result of a 
 successful object audit.
"

^self auditWithLimit: sizeLimit fullChecks: false
%
category: 'Garbage Collection'
method: Repository
auditWithLimit: sizeLimit fullChecks: fullChecking

"Same as auditWithLimit:, except when fullChecking = true, method will 
 return with an error if not in single user mode or if reclaimAll cannot 
 complete.  This guarantees that the object audit will either run using 
 the more detailed consistency checks, or will not run at all.

 Errors returned include:
   #rtErrNotOnlyUser - if not run in single-user mode
   #abortErrReclaimAllFailure - if reclaimAll connot finish

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."

 System needsCommit
   ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _objectAuditWithLimit: sizeLimit fullChecks: fullChecking
%
category: 'Backup and Restore'
method: Repository
commitRestore

"Terminates a restore operation and permits normal commits.

 When executed using Topaz, the result is either a String describing the success
 of the operation (in which case the Topaz result (obj **) may be nil), or an
 error message.

 This method terminates GemStone Smalltalk execution and does an automatic
 logout.  You will need to relogin to continue.

 The restore operation must have been started with restoreFromBackup:.
 Otherwise, this method generates an error.

 If restoreFromCurrentLogs was not the immediately preceding restore operation,
 then a warning is issued, but the termination of restore will succeed.  Such
 use of commitRestore can result in failure to restore all previously committed
 transactions.  However, this allows a Repository to be restored as far as
 practical when some log files are lost or corrupted.

 If GemStone was using partial-logging mode at the time restored backup file(s)
 were written, or if GemStone is currently in partial-logging mode
 (STN_TRAN_FULL_LOGGING is false in the Stone's configuration file), then
 commitRestore is not needed, since the last restoreFromBackup: will have
 committed the restore.

 You must be the only user logged in, otherwise an error is generated.
 The restoreFromBackup: that started the restore process will have
 suspended logins, and a successful commitRestore will reenable logins.

 The session is put into #manualBegin transaction mode, and is left outside
 of a transaction after this method executes.

 This method requires the FileControl privilege.  It is recommended
 that you run as either DataCurator or SystemUser."

^ self _restoreFrom: nil opcode: 2 .
%
category: 'Backup and Restore'
method: Repository
continueFullBackupCompressedTo: fileOrDevice MBytes: mByteLimit

"Same as continueFullBackupTo:MBytes:, except that the output file is
 written compressed in gzip format, and cannot be written to a raw device.
 The output file must be on a local file system; a pgsvr will not be used.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 Backup files written to a file system in compressed mode will have 
 the suffix '.gz' appended to the file name if the specified file name does
 not end with that suffix."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _fullBackupTo: fileOrDevice
    backupId: nil MBytes: mByteLimit
    firstFile: false compressed: true .
%
category: 'Backup and Restore'
method: Repository
continueFullBackupTo: fileOrDevice MBytes: mByteLimit

"Continue a full backup by writing a second or subsequent backup file
 as specified by fileOrDevice, with a size limit specified by mByteLimit.

 This method operates outside of a transaction, and leaves the session
 outside of a transaction.  The session may do one or more aborts during
 the execution of the backup to avoid causing excessive repository growth.

 See fullBackupTo:MBytes: for additional description of the arguments.

 Returns true if the backup was completed.  Returns a message (a String) if
 continueFullBackupTo:MBytes: should be run to complete the backup.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 This method requires the FileControl privilege.

 A GciHardBreak during this method will terminate the session."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _fullBackupTo: fileOrDevice
    backupId: nil MBytes: mByteLimit
    firstFile: false compressed: false .
%
category: 'Copying'
method: Repository
copy

"Disallowed."

self shouldNotImplement: #copy
%
category: 'Extent Operations'
method: Repository
createExtent: extentFilename

"Creates a new Extent with a file named extentFilename (a String).  The new
 Extent has no maximum size.  The extent must be located on the same machine
 running the stone process.  Remote extents are no longer supported.

 If this method is run from a session on a host remote from the stone, 
 extentFilename must include a Network Resource String (NRS) specifying the
 stone host. Otherwise, stone will assume the file is remote and the method
 will fail.
 
 For example, with a stone on 'stonehost' and the gem on 'gemhost': 

   SystemRepository createExtent: '/somedir/ext1.dbf'

 will result in an error, since /somedir/ext1.dbf is assumed to be a directory
 on 'gemhost' (remote from the stone). This is true even if the directory
 specified is a valid directory on 'stonehost'. The proper invocation is:

   SystemRepository createExtent: '!TCP@stonehost#dbf!/somedir/ext1.dbf'
 
 See Appendix C of the GemStone System Administration Guide for information
 on NRS syntax.

 This method updates the DBF_EXTENT_NAMES Stone option.  It does not require
 the system to be in single-user mode.

 If the given file already exists, then this method generates an error and the
 given Extent is not added to the logical Repository.

 Creating an extent with this method bypasses any setting the user may have
 specified for the DBF_PRE_GROW option at system startup.  As extents created
 with this method have no maximum size, their files cannot be grown to an
 initial size.

 If GemStone is being run using weighted disk resource allocation, then
 the new Extent will be given a weight equal to the average weight of all
 other extents.

 Caution:  if you use PSR/PDR GcGems on all extents simultaneously, your 
 maximum session count will be effectively reduced by 1 for each extent 
 created.  The maximum session count will be restored when the PSR/PDR
 GcGems are stopped or when the stone is restarted.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser."

^ self createExtent: extentFilename withMaxSize: -1
%
category: 'Extent Operations'
method: Repository
createExtent: extentFilename withMaxSize: aSmallInteger

"Creates a new Extent with the given extentFilename (aString) and sets the
 maximum size of that Extent to the the given size.  The extent must be 
 located on the same machine running the stone process.  Remote extents are 
 no longer supported. 

 If this method is run from a session on a host remote from the stone, 
 extentFilename must include a Network Resource String (NRS) specifying the
 stone host. Otherwise, stone will assume the file is remote and the method
 will fail.
 
 For example, with a stone on 'stonehost' and the gem on 'gemhost': 

   SystemRepository createExtent: '/somedir/ext1.dbf'

 will result in an error, since /somedir/ext1.dbf is assumed to be a directory
 on 'gemhost' (remote from the stone). This is true even if the directory
 specified is a valid directory on 'stonehost'. The proper invocation is:

   SystemRepository createExtent: '!TCP@stonehost#dbf!/somedir/ext1.dbf'
 
 See Appendix C of the GemStone System Administration Guide for information
 on NRS syntax.

 The size must be a non-zero positive integer representing the maximum physical
 size of the file in megabytes.

 This method updates the DBF_EXTENT_NAMES and DBF_EXTENT_SIZES Stone options.
 It does not require the system to be in single-user mode.

                                  Note:
   The extent is created with the default ownership and permissions of the
   Stone process.  If this is not the same as the ownership and permissions of
   the other extents or replicates of extents, then Unix utilities must be
   used to change the ownership or permissions of the new file; such changes
   may be made without stopping the Stone, and should be made a soon as
   possible, to avoid other sessions encountering authorization errors.

 If the given file already exists, then this method generates an error and the
 given Extent is not added to the logical Repository.

 If the Stone option DBF_PRE_GROW is set to true, then this method will cause
 the newly created extent's file to be grown to the given size.  If the grow
 fails, then this method returns an error and the new Extent is not added to the
 logical Repository.

 If GemStone is being run using weighted disk resource allocation, then
 the new Extent will be given a weight equal to the average weight of all
 other Extents.

 Caution:  if you''re using PSR/PDR GcGems, your maximum session count
 will be temporarily reduced by 2 for each extent created.  The maximum
 session count will be restored when the stone is restarted.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser."

<primitive: 284>

extentFilename _validateClass: String.
aSmallInteger _validateClass: SmallInteger.
^ self _primitiveFailed: #createExtent:
%
category: 'Repository Conversion'
method: Repository
createObjListFrom: fileName ofClass: className

"Private. Returns an Array of (valid) objects based on the list of
 OOPs contained in the text file fileName. The objects must all be of 
 class className."

<primitive: 481>
self _primitiveFailed: #createObjListFrom:ofClass: .
self _uncontinuableError
%
category: 'Repository Conversion'
method: Repository
createObjListFrom: fileName ofKind: className

"Private. Returns an Array of (valid) objects based on the list of
 OOPs contained in the text file fileName. The objects must all be of
 kind className."

<primitive: 478>
self _primitiveFailed: #createObjListFrom:ofKind: .
self _uncontinuableError
%

!edited comments for fix 37873
category: 'Replicate Operations'
method: Repository
createReplicateOf: extentFilename named: replicateFilename

"Create a replicate for the extent with the given name.  (extentFilename and
 replicateFilename are both Strings.)  The replicate must be located on the 
 same machine running the stone process.  Remote replicates are no longer 
 supported.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.  In addition, you must be the only
 user logged into the system.

                                  Note:
   The extent is created with the default ownership and permissions of the
   Stone process.  If this is not the same as the ownership and permissions of
   the other extents or replicates of extents, then Unix utilities must be
   used to change the ownership or permissions of the new file; such changes
   may be made without stopping the Stone.

 Each extent is limited to one replicate.  If the given extent already has a
 replicate, this method generates an error.

 If the given Extent is not a member of the logical Repository, this method
 generates an error.

 If the given replicate file already exists, then this method generates an
 error.

 If any mid-level caches have been created since the stone started,
 even if those caches no longer exist or have no sessions using them,
 this method generates an error.  Use of mid-level caches and replicates
 are mutually exclusive.

 If the Stone option DBF_PRE_GROW is set to true, then this method will cause
 the newly created replicate's file to be grown to the maximum size of the
 extent it mirrors.  If the grow fails, then this method returns an error and
 the new replicate will not be made available.

 The Garbage Collector session is shut down for the duration of this method."

<primitive: 285>

extentFilename _validateClass: String.
replicateFilename _validateClass: String.
^ self _primitiveFailed: #createReplicateOf:named:
%
category: 'Transaction Logging'
method: Repository
currentLogDirectoryId

"Returns a positive SmallInteger, which is the one-based offset specifying the
 element of the configuration list STN_TRAN_LOG_DIRECTORIES for the current
 transaction log.  (See also the currentLogFile method.)"

^ (System stoneConfigurationAt:#StnCurrentTranLogDirId) + 1 
%
category: 'Transaction Logging'
method: Repository
currentLogFile

"Returns a String containing the file name of the transaction log file to which
 log records are being appended.  If the result is of size 0, then a log has
 failed and a replicate is being used."

 ^ (System stoneConfigurationAt:#StnCurrentTranLogNames) at: 1 
%
category: 'Transaction Logging'
method: Repository
currentLogFileId

"Returns a positive SmallInteger, which is the internal fileId of 
 the current transaction log."

^ (self _logInfo: (self currentLogDirectoryId)) at: 4
%
category: 'Transaction Logging'
method: Repository
currentLogReplicate

"Returns a String containing the file name of the transaction log replicate file
 to which log records are being appended. The result is a String of size 0 if
 the current log is not replicated."

 ^ (System stoneConfigurationAt:#StnCurrentTranLogNames) at: 2 
%
category: 'Transaction Logging'
method: Repository
currentTranlogSizeMB

"Returns an Integer that is the size of the currently active transaction log in
 units of Megabytes."

^ self _currentLogInfo at: 5
%
category: 'Accessing'
method: Repository
dataDictionary

"Accesses the user-defined data dictionary.  (Not for use
 in this product release.)"

^ dataDictionary
%
category: 'Replicate Operations'
method: Repository
disposeReplicate: replicateFilename

"Removes the given replicate from, whose name is specifies as aString, from the
 system.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.  In addition, you must be the only
 user logged into the system.

 If the given replicate is not a real Repository replicate, this method
 generates an error.

 The Garbage Collector session is shut down for the duration of this method."

<primitive: 287>

replicateFilename _validateClass: String.
^ self _primitiveFailed: #disposeReplicate:
%
category: 'Clustering'
method: Repository
extentForPage: aPageId

"Returns a SmallInteger specifying an offset into the result from the
 fileNames method.  

 The argument aPageId is an Integer, such as the result from the Object | page
 method, specifying a page in the receiver."

^ (aPageId // 16#7FFFFF) + 1
%
category: 'Accessing'
method: Repository
fileNames

  <primitive: 429>

"Returns an Array of Arrays containing the filenames for the extents and
 replicates of the receiver.

 Each Array within the returned Array contains two Strings.  The first String
 represents the filename of the Nth extent (where N is the index into the
 returned Array).  The second element represents the filename of the replicate
 of the Nth extent.  If there is no replicate, the 2nd element is a zero-length
 String."

  self _primitiveFailed: #fileNames.
  self _uncontinuableError
%
category: 'Repository Usage Reporting'
method: Repository
fileSize

"Returns an integer giving the total physical size, in bytes, of all
 of the physical extents that compose the logical Repository."

| total |

total := 0.
1 to: self numberOfExtents do:[:j |
   total := total + ((self _extentStatistics: j) at: 1)
   ].
^total
%
category: 'Repository Usage Reporting'
method: Repository
fileSizeOfExtent: extentFilename

"Returns the physical size, in bytes, of the extent with the given name.

 If the given file is not an active member of the logical Repository
 represented by the receiver, then this method generates an error."

^(self _dbfStatistics: extentFilename) at: 1
%
category: 'Repository Usage Reporting'
method: Repository
fileSizeReport

"Returns a string which reports on the name, size, and amount of free space
 for each extent and the size and free space of the entire logical Repository."

| result lf stats repositoryFiles extentFiles |

    "Initialize"

result := String new.
lf := Character lf.

    "For each extent, add the information about it."

repositoryFiles := self fileNames.
1 to: self numberOfExtents do: [:i |
    stats := self _extentStatistics: i.
    extentFiles := repositoryFiles at: i.
    result addAll: 'Extent #';
           addAll: i asString; add: lf.
    result addAll: '-----------'; add: lf.
    result addAll: '   Filename = ';
           addAll: (extentFiles at: 1); add: lf.
    result addAll: '   Replicate = '.
    (extentFiles at: 2) size == 0
       ifTrue: [ result addAll: 'NONE' ]
       ifFalse: [ result addAll: (extentFiles at: 2) ].
    result addAll: lf  ;
           addAll: '   File size =       ';
           addAll: (self numToMByteString:(stats at: 1)) ; add: lf;
           addAll: '   Space available = ';
           addAll: (self numToMByteString:(stats at: 2)) ; add: lf;
           add: lf .
    ].

    "Add totals"

result addAll: 'Totals'; add: lf.
result addAll: '------'; add: lf.
result addAll: '   Repository size = ';
       addAll: (self numToMByteString: self fileSize) ; add: lf;
       addAll: '   Free Space =      ';
       addAll: (self numToMByteString: self freeSpace ) ; add: lf .

^result
%

category: 'Garbage Collection'
method: Repository
findDisconnectedObjects

"This method helps determine the kinds of objects that are dead in GemStone and
 thus allows you to tune your code to help prevent unnecessary objects from
 being stored on disk.

 This method returns an Array containing a list of (non-private) objects that
 are not transitively connected to any permanent object, based on the GemStone
 state as viewed by the transaction in which this method is executed.  These
 objects are considered possibly dead and could be reclaimed by execution of a
 markForCollection.

 It is important that the application disconnect the returned Array after
 examining it to avoid having these possible dead objects be inadvertently
 connected to GemStone permanently.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 This method requires the GarbageCollection privilege."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self _findDisconnectedObjects
%

category: 'Garbage Collection'
method: Repository
_findDisconnectedObjects

"This method helps determine the kinds of objects that are dead in GemStone and
 thus allows you to tune your code to help prevent unnecessary objects from
 being stored on disk.

 This method returns an Array containing a list of (non-private) objects that
 are not transitively connected to any permanent object, based on the GemStone
 state as viewed by the transaction in which this method is executed.  These
 objects are considered possibly dead and could be reclaimed by execution of a
 markForCollection.

 It is important that the application disconnect the returned Array after
 examining it to avoid having these possible dead objects be inadvertently
 connected to GemStone permanently.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 This method requires the GarbageCollection privilege."

<primitive: 430>

self _primitiveFailed: #_findDisconnectedObjs.
self _uncontinuableError
%
category: 'Garbage Collection'
method: Repository
findObjsConnectedTo: anArray
"Return an array of all objects reachable starting from the objects
 in the given array.  Objects implicitly referenced from the
 header such as the object's class and segment are not included in the
 scan.

 This method requires the GarbageCollection privilege.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."    

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

anArray size == 0 ifTrue:[ anArray _error: #rtErrArgOutOfRange ] .
System _hiddenSetReinit: 26.
System _addAll: anArray to: 26.
^self _findConnectedObjs
%
category: 'Repository Usage Reporting'
method: Repository
freeSpace

"Returns an integer that gives the number of bytes of free space in the logical
 Repository.  This quantity is equal to the number of free pages in the
 Repository times the size of a page."

| total |

total := 0.
1 to: self numberOfExtents do:[:j |
   total := total + ((self _extentStatistics: j) at: 2)
   ].
^total
%
category: 'Repository Usage Reporting'
method: Repository
freeSpaceInExtent: extentFilename

"Returns the number of bytes of free space in the extent with the given name.
 This quantity is equal to the number of free pages in the extent times the
 size of a page.

 If the given file is not an active member of the logical Repository
 represented by the receiver, then this method generates an error."

^(self _dbfStatistics: extentFilename) at: 2
%
category: 'Backup and Restore'
method: Repository
fullBackupCompressedTo: fileOrDevice

"Backup the receiver to a single backup file or tape in gzip format.  See
 fullBackupCompressedTo:MBytes: for further documentation.

 Backup files written to a file system in compressed mode will have 
 the suffix '.gz' appended to the file name if the specified file name does
 not end with that suffix."

^ self fullBackupTo: fileOrDevice MBytes: 0 compressed: true
%
category: 'Backup and Restore'
method: Repository
fullBackupCompressedTo: fileOrDevice MBytes: mByteLimit

"Same as fullBackupTo:MBytes: , except that the output file is 
 written compressed in gzip format, and cannot be written to a raw device.
 The output file must be on a local file system; a pgsvr will not be used.

 Backup files written to a file system in compressed mode will have 
 the suffix '.gz' appended to the file name if the specified file name does
 not end with that suffix."

  ^ self fullBackupTo: fileOrDevice MBytes: mByteLimit compressed: true
%
category: 'Backup and Restore'
method: Repository
fullBackupTo: fileOrDevice

"Backup the receiver to a single backup file or tape.  See
 fullBackupTo:MBytes: for further documentation.

 GemStone Systems has QA tested backup/restore of backup files
 up to 16GB.  For larger repositories, we recommend you use
 fullBackupTo:MBytes: to break the backup into several files.

 This method requires the FileControl privilege.

 A GciHardBreak during this method will terminate the session."

^ self fullBackupTo: fileOrDevice MBytes: 0
%
category: 'Backup and Restore'
method: Repository
fullBackupTo: fileOrDevice MBytes: mByteLimit

"Produces a full backup file containing the most recently committed version of
 the receiver as of the time the method is executed.

 The argument fileOrDevice (a kind of String) specifies the file or device
 where the backup is to be created.  

 fileOrDevice can be either a file on some file system, or a device name 
 specifying a tape device.  Backups to raw partitions are no longer 
 supported.

 The fileOrDevice argument may use GemStone Network Resource String syntax.  
 For example, this may be used to access a tape device on another machine, 
 provided a GemStone NetLDI process is running on the remote machine.  

 If fileOrDevice specifies a file that already exists on a fileSystem, an
 error is generated.

 The mByteLimit argument, which specifies the maximum size of fileOrDevice in
 units of megabytes, must be a SmallInteger.  The value 0 means that there is
 no limit on the size of the resulting fileOrDevice.  A mByteLimit less than 0
 or greater than 4096000 will generate an error.  GemStone Systems has QA tested
 backup/restore of backup files up to 16GB.  While larger backup files will
 probably work, they have not been QA tested, so we recommend you limit backup
 file size to 16GB, or conduct your own tests with larger sizes before trusting
 the backup.

 If the backup requires more bytes than you specified in mByteLimit, this method
 returns a message (a String) stating that a partial backup file was created.
 In this case, further commits in this session are disallowed until
 you either complete the full backup with continueFullBackupTo:MBytes: or
 cancel the backup with abortFullBackup.  To continue the backup, you can
 execute the method continueFullBackupTo:MBytes:, which creates the next file
 in the backup sequence.

 If fileOrDevice runs out of space, such as off the end of a tape, the backup
 will terminate with a system I/O error at that point.  The backup will be
 unusable.  To avoid having to repeat the entire backup, make sure the device
 has sufficient space or set mByteLimit appropriately.

 When the size of your GemStone repository exceeds the capacity of a backup
 tape or file system, you can use mByteLimit (a SmallInteger) to control the 
 maximum number of bytes to be written to the backup file.  

 This method puts the session into auto-begin transaction mode and aborts
 the current transaction.  If an abort would cause unsaved  changes to be
 lost, it signals an error, #rtErrAbortWouldLoseData.   If the repository
 is not in restore state, then this method commits a record of the start
 of the backup to UserGlobals at: #BackupLog;  this commit is done as a
 checkpoint.   Then the transaction mode is changed to manual begin, 
 and the remainder of the backup operation executes outside of a
 transaction so that it does not cause excessive repository growth.  
 A varying number of aborts are done while outside of a transaction, 
 depending on the time required to execute the backup.

 When the backup completes, the session is always left outside of a transaction
 so that it does not retain a commit record that would cause the repository to
 grow.

 Returns true if the backup was completed.  Returns a message (a String) if
 continueFullBackupTo:MBytes: should be run to complete the backup.

 This method requires the FileControl privilege.

 A GciHardBreak during this method will terminate the session."

  ^ self fullBackupTo: fileOrDevice MBytes: mByteLimit compressed: false
%
category: 'Private'
method: Repository
fullBackupTo: fileOrDevice MBytes: mByteLimit compressed: comprBool

| result |

 System needsCommit
   ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

(SystemRepository _restoreStatusInfo at: 2) == nil ifTrue:[
   | logObj uGlobals |

   System inTransaction 
     ifFalse:[System beginTransaction ] .  "abort/begin the current transaction"
  "Repository is not in restore state, so record the start of 
   the backup in the log object"
  uGlobals := (GsSession currentSession resolveSymbol: #UserGlobals) value .
  logObj := uGlobals at:#BackupLog ifAbsent:[
    uGlobals at:#BackupLog put: String new 
    ] .
  logObj add: 'fullBackup to ' + fileOrDevice + ' started at ';
       add: (DateTime now asStringUsingFormat: #[2,1,3,$/,2,1,$:,true,false,false]);
       add: Character lf .
  "with tranlog/replay of dead, commit here, and abort to checkpoint in C."
  System commitTransaction 
    ifFalse:[ self _error: #rtErrMaxCommitRetry ] .
].

System inTransaction 
  ifFalse:[System beginTransaction ] .  "primitive must be called within a transaction"

result := self _fullBackupTo: fileOrDevice
               backupId: nil 
               MBytes: mByteLimit 
               firstFile: true 
               compressed: comprBool .

^result
%
category: 'Adding'
method: Repository
insert: aSequenceableCollection at: anIndex

"Disallowed."

self shouldNotImplement: #insert:at:
%
category: 'Adding'
method: Repository
insertAll: aSequenceableCollection at: anIndex

"Disallowed."

self shouldNotImplement: #insertAll:at:
%
category: 'Class Management'
method: Repository
listInstances: anArray

"Returns a list of instances on the receiver that belong to one of the
 classes listed in anArray.  The result of this method is an Array of IdentitySets,
 where the contents of each set consists of all instances whose class is equal
 to the corresponding element in anArray.

                            Warning:
 You may retrieve instances to which you have no read access, so this method
 is mostly of use to SystemUser.

 If anArray contains multiple occurrences of a class, then the result
 will contain corresponding multiple occurrences of the same Set that lists
 the instances of that class.

 If anArray contains an element that is not a kind of Behavior, an error
 is generated.

 Scans the entire Repository at least once.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 If the argument anArray contains more than 2000 unique elements then
 the entire Repository will be scanned once for each group of 2000 unique
 elements, or fraction thereof."

| inputSet inputSetSize inputIdx resultInSetOrder result inputArraySize
  anObj scanSetThisTime |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

inputSet := IdentitySet withAll: anArray .
inputSetSize := inputSet size .
inputSetSize < 1 ifTrue:[ ^ Array new ] .
inputSet do: [:each | each _validateClass: Behavior].

inputIdx := 1 .
resultInSetOrder := Array new .
[inputIdx <= inputSetSize] whileTrue:[
   scanSetThisTime := IdentitySet new .
   [ scanSetThisTime size < 2000 _and:[inputIdx <= inputSetSize ]] whileTrue:[
      scanSetThisTime add: (inputSet _at: inputIdx) .
      inputIdx := inputIdx + 1 .
      ].
   resultInSetOrder addAll: (self _listInstances: scanSetThisTime) .
   ].
inputArraySize := anArray size .
result := Array new: inputArraySize .
1 to: inputArraySize do:[:j|
  anObj := anArray at: j .
  result at: j put: ( resultInSetOrder at: (inputSet _offsetOf: anObj))
  ].
^ result
%

! Request 35612
category: 'Class Management'
method: Repository
listObjectsInSegments: anArray

"Returns a list of committed objects that belong to one of the segments listed
 in anArray.  The result of this method is an Array of IdentitySets, where 
 the contents of each set consists of all instances whose Segment is 
 the corresponding element in anArray.  Dead objects (those in the stone's
 deadNotReclaimed set) are not included in the scan results.

                            Warning:
 You may retrieve instances to which you have no read access, so this method
 is mostly of use to SystemUser.

 If anArray contains multiple occurrences of a class, then the result
 will contain corresponding multiple occurrences of the same Set that lists
 the instances of that class.

 If anArray contains an element that is not a kind of Behavior, an error
 is generated.

 Scans the entire Repository at least once.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 If the argument anArray contains more than 2000 unique elements then
 the entire Repository will be scanned once for each group of 2000 unique
 elements, or fraction thereof."

| inputSet inputSetSize inputIdx resultInSetOrder result inputArraySize
  anObj scanSetThisTime |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

inputSet := IdentitySet withAll: anArray .
inputSetSize := inputSet size .
inputSetSize < 1 ifTrue:[ ^ Array new ] .
inputSet do: [:each | each _validateClass: Segment].

inputIdx := 1 .
resultInSetOrder := Array new .
[inputIdx <= inputSetSize] whileTrue:[
   scanSetThisTime := IdentitySet new .
   [ scanSetThisTime size < 2000 _and:[inputIdx <= inputSetSize ]] whileTrue:[
      scanSetThisTime add: (inputSet _at: inputIdx) .
      inputIdx := inputIdx + 1 .
      ].
   resultInSetOrder addAll: (self _listObjectsInSegments: scanSetThisTime) .
   ].
inputArraySize := anArray size .
result := Array new: inputArraySize .
1 to: inputArraySize do:[:j|
  anObj := anArray at: j .
  result at: j put: ( resultInSetOrder at: (inputSet _offsetOf: anObj))
  ].
^ result
%

category: 'Class Management'
method: Repository
listReferences: anArray 

"Returns a list of instances in the receiver that have a reference to one of
 the objects specified in anArray.  The result of this method is an Array of
 sets, where the contents of each set consists of all instances that have a
 reference to the corresponding element in anArray.  Instances to which you
 have no read authorization are silently omitted.

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method is very expensive.  It scans the entire Repository and looks
 at every instance variable of every object."

| result |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

result := self _listReferences: anArray withLimit: 0.

"Remove the first slot in the arrays which indicate whether an authorization
 error occurred or not."
1 to: result size do: [ :i | | oldArray newArray |
  oldArray := result at: i.
  newArray := oldArray copyFrom: 2 to: (oldArray size).
  result at: i put: newArray.
  ].
^ result.
%
category: 'Class Management'
method: Repository
listReferences: anArray withLimit: aSmallInt

"Returns a list of instances in the receiver that have a reference to one of
 the objects specified in anArray.  The result of this method is an Array of
 Arrays, where the contents of each Array consists of instances that have a
 reference to the corresponding element in anArray.  The number in each Array
 is limited by aSmallInt.  If an instance for which you have no read
 authorization is found, the result Array contains nil.

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method is very expensive.  It scans the entire Repository and looks
 at every instance variable of every object."

| result |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

result := self _listReferences: anArray withLimit: aSmallInt.

1 to: result size do: [ :i | | oldArray |
  oldArray := result at: i.
  (oldArray at: 1) 
    ifTrue:  [ oldArray at: 1 put: nil ]
    ifFalse: [ | newArray |
      newArray := oldArray copyFrom: 2 to: (oldArray size).
      result at: i put: newArray.
      ].
  ].
^ result.
%
category: 'Transaction Logging'
method: Repository
logOriginTime

"Returns the log origin time of the receiver. This is the time when a Stone
 started a new sequence of log files for the receiver. A new sequence of
 logs is started if one of the following occurs:

 * Stone is started using extents that were cleanly shutdown, and without
   any log files being present.

 * Stone is started using extents and no existing logs 
   using the 'startstone -N' command.

 * The commitRestore method is executed, and during preceding 
   restore operations we restored a log file with fileId greater
   than the fileId of the log file being written to during the restore."

^ System stoneConfigurationAt:#StnTranLogOriginTime 
%
category: 'Garbage Collection'
method: Repository
markForCollection

"Performs a garbage collection analysis of all permanent objects on disk.  
 Every object in the receiver that cannot be reached from AllUsers is marked 
 for subsequent reclaiming of storage space. 

 This method aborts the current transaction, empties the GcCandidates queue 
 and commits, runs the analysis while outside of a transaction and then reenters
 a transaction if the session was in a transaction at the start of this method.

 When this method completes successfully, it generates an error message of
 the following form:

   Successful completion of Garbage Collection
     found <anInt> live objects,
     found <anInt> dead objects, occupying <anInt> bytes

 After this method completes, the GcGem process automatically reclaims the space
 occupied by the dead objects.  Space is not reclaimed until other sessions have
 either committed or aborted the transactions that were concurrent with this
 method.

 This method may fail with an error if another garbage collection is
 in progress.  This method waits up to 2 minutes for the other garbage 
 collection to finish before issuing the error.
 Another garbage collection can be in progress because:
   1) an epoch garbage collection is in progress in a gc gem.
   2) a markGcCandidates is in progress in another session
   3) a markForCollection is in progress in another session
   4) a previous epoch, markGcCandidates or markForCollection completed the
     mark phase, but the voting on the possibly dead objects has 
     not completed. Note that a gc gem must be running for the 
     voting to complete.  Also, long running sessions that are idle and 
     never aborting will prevent voting from completing. 

 This method requires the GarbageCollection privilege.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 A GciHardBreak during this method terminates the session."

^ self markForCollectionWait: 120 

%

category: 'Garbage Collection'
method: Repository
markForCollectionWait: waitTimeSeconds

"Perform a markForCollection, waiting up to waitTimeSeconds for any
 garbage collection in progress in another session to complete.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 To wait as long as necessary for the other gc to finish, pass the argument -1.
 This should be done with caution, however, as under certain conditions
 the session could appear to wait forever.  To avoid this you need to:

 1.  Make sure that other sessions are committing/aborting to allow
     voting on possible dead to complete.

 2.  Make sure that either the normal GcGem or the DeadReclaim (DR) GcGem
     is running to complete processing of dead objects once the vote is
     completed.

 For further documentation see markForCollection. "

self prepareForMarkForCollection .
^self _primMarkForCollection: waitTimeSeconds
%


category: 'Garbage Collection'
method: Repository
prepareForMarkForCollection

"Prepares to run a markForCollection operation."

| candidates candidatesQueue candidatesCount |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

System abortTransaction.
(System transactionMode == #manualBegin) ifTrue: [System beginTransaction].

candidatesQueue := Globals at: #GcCandidates.
candidatesCount := Globals at: #GcCandidatesCount.

candidates := candidatesQueue removeAll.
candidatesQueue cleanupQueue.
candidatesCount decrementBy: (candidatesCount value).
candidates size: 0.
System commitTransaction ifFalse: [
  | errorNumber |
  System abortTransaction.
  errorNumber := self _errorSymbolToNumber: #abortErrGarbageCollection args: #[].
  System signal: errorNumber args: #['Commit failed after GcCandidates cleanup', nil] signalDictionary: GemStoneError
  ].
self _buildWeakRefSet.
^true
%

category: 'Garbage Collection'
method: Repository
prepareForMarkGcCandidates

"Prepares to run a markGcCandidates operation."

| candidates anArray element candidatesQueue candidatesCount|

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

System abortTransaction.
(System transactionMode == #manualBegin) 
  ifTrue: [System beginTransaction].
candidatesQueue := Globals at: #GcCandidates.
candidatesCount := Globals at: #GcCandidatesCount.
candidates := candidatesQueue removeAll.
candidatesQueue cleanupQueue.
candidatesCount decrementBy: (candidatesCount value).
1 to: candidates size do: [:i |
  anArray := candidates at: i.
  1 to: anArray size do: [:j |
    element := anArray at: j.
    (element isSpecial) "ignore special objects"
       ifFalse: [System _add: (anArray at: j) to: 34]
    ].
  System _add: anArray to: 34.
  anArray size: 0.
  ].
candidates size: 0.
System commitTransaction ifFalse: [
  System abortTransaction.
  self _error: #abortErrGarbageCollection args: #[]
  ].

^ true
%

category: 'Garbage Collection'
method: Repository
markGcCandidates

"Performs a garbage collection analysis on the objects in GcCandidates by
 scanning the Repository for other references to them.  Objects are marked for
 subsequent reclaiming of storage space if the only references are from other
 candidates.

 Use the Repository | addGcCandidates: method to add elements to the queue.
   
 This method may fail with an error if epoch garbage collection or 
 markForCollection is in progress at the time it is executed.  In this case,
 markGcCandidates should be retried after waiting a few minutes.

 This method empties the GcCandidates queue into a temporary Array and
 commits. 
     
 When this method completes successfully, it returns an Array containing
 at least two elements:

    Array[1] = number of possible dead Objects found.
    Array[2] = number of bytes in the possible dead objects.

 If there are entries in the Array after the second then:
    Array[3..objSize] = the entries in the GcCandidates queue that
               were not eligible for collection.
                    
 Space for the objects determined to be dead is not reclaimed until other
 sessions have either committed or aborted the transactions that were
 concurrent with this method.
   
 This method requires the GarbageCollection privilege.
    
 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 A GciHardBreak during this method will terminate the session."

self prepareForMarkGcCandidates .
^ self _markGcCandidates
%
category: 'Accessing'
method: Repository
name

"Returns the logical name of the receiver (a Symbol)."

^ name
%
category: 'Updating'
method: Repository
name: aString

"Redefines the logical name of the receiver to be aString."

name := aString asSymbol
%
category: 'Extent Operations'
method: Repository
numberOfExtents

  <primitive: 428>

"Returns the number of active extents."

  self _primitiveFailed: #numberOfExtents.
  self _uncontinuableError
%
category: 'Repository Usage Reporting'
method: Repository
numToMByteString: aNumber

"Convert a number representing a file size in bytes to a formatted
 string reporting the size in megabytes."

| result |
result := String new .
result addAll:
  ((aNumber asFloat / 1048576) asStringUsingFormat:#(3 2 false));
  addAll: ' Megabytes' .
^ result
%
category: 'Garbage Collection'
method: Repository
objectAudit

"Checks all objects in GemStone for consistency.

 This method is equivalent to the message auditWithLimit: 100000.

 See objectAuditFullChecks for a variation that runs in single-user mode.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 See auditWithLimit: for further documentation."

^ self auditWithLimit: 100000
%
category: 'Garbage Collection'
method: Repository
objectAuditFullChecks

"Same as objectAudit, except method will return with an error if not in
 single user mode or if reclaimAll cannot complete.  This guarantees that
 the object audit will either run using the more detailed consistency checks,
 or will not run at all.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 Errors returned include:
   #rtErrNotOnlyUser - if not run in single-user mode
   #abortErrReclaimAllFailure - if reclaimAll cannot finish"
  
^ self auditWithLimit: 100000 fullChecks: true
%
category: 'Transaction Logging'
method: Repository
oldestLogFileIdForRecovery

"Returns a positive SmallInteger, which is the internal fileId of the oldest
 transaction log needed to recover from the most recent checkpoint, if the
 Stone were to fail right now."

^ self _currentLogInfo at: 10
%
category: 'Repository Usage Reporting'
method: Repository
pageSize

"Returns size in bytes of a disk page in the Repository."

^ 8192
%
category: 'Garbage Collection'
method: Repository
pagesWithPercentFree: aPercent

"This method returns an Array containing the following statistics:

 1.  The total number of data pages processed.
 2.  The sum in bytes of unused space in all data pages.  This quantity is a
     measure of data fragmentation in the receiver.
 3.  The number of bytes in a page.
 4.  The number of data pages that have at least the specified percentage of
     unused space.

 Do not confuse unused space on a page with free space (unused pages) in a
 Repository or Extent.  See the freeSpace and freeSpaceInExtent: methods
 for more information.

 This method requires the GarbageCollection privilege.

 A GciHardBreak during this method will terminate the session."

^ self _pagesWithPercentFree: aPercent doScavenge: false
%
category: 'Formatting'
method: Repository
printOn: aStream

"Puts a displayable representation of the receiver on the given stream."

"Copy the implementation from Object so we don't inherit it from Collection."

aStream nextPutAll: self asString
%
category: 'Garbage Collection'
method: Repository
quickObjectAuditLevel1

"Performs a level 1 quick object audit of the repository.  Fails if not run in 
 single user mode (except for GcGem, which is stopped automatically).  Object
 statistics are not reported in the interest of attaining fastest performance.

 The quick audit first requests the stone reclaim all shadow pages and dead 
 objects.  This part of the audit may run for long periods of time if many 
 shadow pages and/or dead objects exist in the repository. 

 The quick audit will fail with an error if possible dead object finalization is in
 progress.  If this happens, ensure the GcGem is running and try the quick audit
 again later.  To prevent this error condition, do not run the quick audit
 immediately after a markForCollection, markGcCandidates, or epoch GC operation
 has completed.

 The object table is not audited by this method.  Most other checks done in the
 standard object audit (Repository>>auditWithLimit:) are performed.  References to
 any non-existent, free or dead oops will be reported as errors. 

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 This method will run faster on large repositories than the
 Repository>>auditWithLimit: method."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self quickObjectAuditWithLevel: 1.
%
category: 'Garbage Collection'
method: Repository
quickObjectAuditLevel2

"Performs a level 2 quick object audit of the repository.  Performs the same
 integrity audits as Repository>>quickObjectAuditLevel1.  In addition, all object
 table entries are audited to verify the disk address of each object is correct.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 This method will longer to complete than Repository>>quickObjectAuditLevel1 on
 large repositories."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self quickObjectAuditWithLevel: 2.
%
category: 'Private'
method: Repository
quickObjectAuditWithLevel: anInt

<primitive: 524>

self _primitiveFailed: #quickObjectAuditWithLevel:.
self _uncontinuableError
%
category: 'Garbage Collection'
method: Repository
reclaimAll

"Explicitly triggers the reclamation of all shadowed and dead objects.
 Returns true if the reclaim process completed normally, otherwise returns
 false.  Successful completion requires this session be the only non-gc 
 session logged in, the right combination of GcGems be running, voting on 
 dead objects must be completed, and the GcUser configuration parameter
 reclaimDeadEnabled be true.  Valid GcGem combinations are: [none],
 [GcGem only], [RCL, EPC].

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method requires the GarbageCollection privilege."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^System _reclaimAll.
%
category: 'Garbage Collection'
method: Repository
repairWithLimit: sizeLimit

"Checks all objects in GemStone for consistency and repairs any errors found.
 A description of errors found and repaired is written to standard
 output, along with statistics about the Repository.  The statistics report does
 not include any objects smaller than the specified sizeLimit (number of bytes
 or OOPs).

 This method should be executed from topaz -l (the linked version of Topaz).

 This method requires the GarbageCollection privilege.  In addition, you must
 be the only user logged into GemStone.

 This method aborts the current transaction.

 The Garbage Collector session is shut down for the duration of this method.

 A GciHardBreak during this method will terminate the session.

 If this session is the only session logged, logins are disabled for the
 duration of this method, scavenging is forced to complete, and additional
 consistency checks are made during the audit/repair.   If other users were 
 found to be logged in, then scavenging is not completed, logins are not 
 disabled, and the audit/repair performs less checking of the Repository.

 If for some reason scavenging cannot be completed, this method runs as if
 other users were logged in, and the audit/repair performs less checking of 
 the Repository. Refer to repairWithLimit:fullChecks: for a variation that 
 will return immediately with errors rather than performing the audit/repair
 with fewer checks.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method raises the special errors #3021 or #3032.  Topaz performs 
 special processing over these results to determine whether or not the audit 
 succeeded.  The Topaz  'expectvalue true' command matches the result of a 
 successful object repair.
"

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self repairWithLimit: sizeLimit fullChecks: false
%
category: 'Garbage Collection'
method: Repository
repairWithLimit: sizeLimit fullChecks: fullChecking

"Same as repairWithLimit:, except when fullChecking = true, method will 
 return with an error if not in single user mode or if reclaimAll cannot 
 complete.  This guarantees that the audit/repair will either run using 
 the more detailed consistency checks, or will not run at all.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.   

 Errors returned include:
   #rtErrNotOnlyUser - if not run in single-user mode
   #abortErrReclaimAllFailure - if reclaimAll connot finish"

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _objectRepairWithLimit: sizeLimit destroyIndexes: false fullChecks: fullChecking
%
category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogDirectories: arrayOfDirectorySpecs
tranlogPrefix: tranlogPrefixString
replicateDirectories: arrayOfReplicateDirSpecs
replicatePrefix:  replicPrefixString

"This method is equivalent to invoking 
 setArchiveLogDirectories:tranlogPrefix:replicateDirectories:replicatePrefix:
 followed by restoreFromArchiveLogs.

 Please see those two methods for complete descriptions."

self setArchiveLogDirectories: arrayOfDirectorySpecs
  tranlogPrefix: tranlogPrefixString
  replicateDirectories: arrayOfReplicateDirSpecs
  replicatePrefix:  replicPrefixString .

^ self restoreFromArchiveLogs
%
category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs

"Given a Repository already in restore mode from a previous restore operation,
 restores all available transaction logs contained in the directories specified
 by the last preceding invocation of either 

    Repository>>setArchiveLogDirectories:...replicatePrefix:   

  or

    Repository>>restoreFromArchiveLogDirectories:...replicatePrefix:   

 Determines the restore status's current fileId by doing the equivalent of
 SystemRepository restoreStatus.  Then attempts to restore contents of any log
 file whose current fileId is beyond the end of the last restore, if the log
 file can be found when searching the directories previously specified.

 Generates an error if neither setArchiveLogDirectories:... nor
 restoreFromArchiveLogDirectories:... has been executed since the last 
 startstone of this Repository.

 When opening a log file on a file system, if the filename ends in '.gz',
 the file is expected to be in gzip compressed format, and will be
 opened as such.  Restoring a compressed tranlog from a raw 
 partition is not supported.

 When executed using Topaz, the result is either a String describing the success
 of the operation (in which case the Topaz result (obj **) may be nil), or an
 error message.

 This method terminates GemStone Smalltalk execution and does an automatic
 logout.  You will need to relogin to continue.

 You must be the only user logged in, otherwise an error is generated.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.  This method puts the session into
 #manualBegin transaction mode.

 Note that restore status is an attribute of the Repository, not of a
 session, so the required preceding restore operation could have been
 executed in some preceding session."

System transactionMode: #manualBegin .
System abortTransaction .

^ self _restoreFrom: nil opcode: 9
%
category: 'Backup and Restore'
method: Repository
restoreFromBackup: fileOrDevice

"If a restore is not in progress, starts a full restore of the receiver by
 initializing a shadow object space and reading the first backup file into that
 space.  Normal commits are disallowed while a restore is in progress.

 If a restore is in progress, continues the restore by reading a second or
 subsequent backup file from a multi-file backup set.

 Use the method restoreStatus to determine whether a restore is in progress
 or not, and the next file expected in a multiple file restore.   Use the
 method abortRestore to cancel a restore that stopped prematurely due to
 fileOrDevice being truncated or corrupt, before attempting the restore
 with a good copy of fileOrDevice.

 fileOrDevice can be either a file on some file system, or a device name 
 specifying a tape device.  Restores from raw partitions are no longer 
 supported.

 If the fileOrDevice is the last backup file in a backup set, the shadow object
 space is automatically made visible to GemStone Smalltalk at the completion of
 this method, and if GemStone was in full-logging mode at the time of the
 backup, the object server is made ready for restoreFromLog:.  This installation
 of the restored object table terminates GemStone Smalltalk execution and does
 an automatic abort.  All GemStone Smalltalk temporary objects present at the
 start of this method are destroyed, so this method can only be executed from
 Topaz.  Once the last file backup file in a backup set has been restored, the
 restore status of the Repository will persist across sessions and shutdowns of
 the Stone.

 If the fileOrDevice is the last backup file in a backup set and GemStone
 was in partial logging mode at the time of the backup, then the Repository
 is ready for normal use after the restore of the file.

 If the last file of a backup set has not yet been restored, the shadow object
 space is thrown away if this session logs out, or if abortRestore is executed.
 After a fresh login the restore would have to be restarted with the first
 backup file again.

 After restoring the last file in a series of backups, this method terminates
 GemStone Smalltalk execution and does an automatic logout and you need
 to relogin to continue.   If it is not the last file in the backup, then 
 the result is either a String  describing the success of the operation 
 (in which case the Topaz result (obj **)  may be nil), or an error message.  

 The backup file must have been previously created with one of the following:
 fullBackupTo:,  fullBackupTo:MBytes:, continueFullBackupTo:MBytes:,
 fullBackupCompressedTo:, fullBackupCompressedTo:MBytes:, or
 continueFullBackupCompressedTo:MBytes: .  The restore operation will
 automatically attempt to open a file in compressed mode with and without
 appending a '.gz' suffix if reading the file as uncompressed does not
 succeed.

 Restored objects will be clustered in a manner that is similar, but not
 necessarily identical, to the clustering organization at the time the backup
 file was created.  

 If the Repository being restored into has the same number of extents as the
 Repository from which the full backup was made, then distribution of objects
 within extents is preserved.  In this case the DBF_ALLOCATION_MODE
 configuration parameter is ignored during the restore, unless an extent hits a
 size limit specified by DBF_EXTENT_SIZES.  If the number of extents differs,
 then the DBF_ALLOCATION_MODE configuration parameter at the time of the restore
 will control distribution of objects across the extents.  The number of extents
 recorded in the backup file is the number of extents as of the start of the
 full backup.   

 You must be the only user logged in, otherwise an error is generated.  This
 method suspends logins.  Logins will be reenabled when one of the following
 occurs:

 1) this session logs out.
 2) the last backup files(s) of a backup have been restored, and the backup was
    made when in partial logging mode.
 3) commitRestore succeeds.

 It is recommended that the Stone be restarted on a copy of the initial
 Repository, $GEMSTONE/bin/*.dbf, before executing this method, in order to
 minimize the size of the restored Repository.

 The Garbage Collector session is shut down at the beginning of this method.
 If this method succeeds, then the Garbage Collector session remains shut down
 until restoreFromCurrentLogs has been successfully executed, otherwise
 the Garbage Collector session may be scheduled for restarting.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.

 A GciHardBreak during this method will terminate the session.

 Only sessions using a shared page cache may restore from backup.

 If the session is using a shared page cache, then the asynchronous I/O function
 of the Stone's page server process is made more aggressive.  The following
 settings are automatically active for the duration of this method:

   System configurationAt:#StnMntMaxAioRate put: 1000 .
   System configurationAt:#ShrPcTargetPercentDirty put: 5 ."

^ self _restoreFrom: fileOrDevice opcode: 0
%

category: 'Backup and Restore'
method: Repository
restoreFromBackup: fileOrDevice newSystemUserPassword: aString

"Same as the restoreFromBackup: method except that a new password is
 set for SystemUser in the restored repository.  Passwords for all
 other user profiles will be restored to their values at the time 
 the backup was taken.

 If aString is nil, then this method functions the same as the
 restoreFromBackup: method and the SystemUser password is not
 changed.

 This method must be run as SysterUser if a new SysterUser password is
 specified."

^ self _restoreFrom: fileOrDevice opcode: 0 newSystemUserPassword: aString
%

category: 'Backup and Restore'
method: Repository
restoreFromBackups: arrayOfFilesOrDevices

"Restore multiple backup files.  This method is designed to perform faster
 than executing restoreFromBackup: for each element of arrayOfFilesOrDevices.  

 The arrayOfFilesOrDevices argument must be an Array not larger than 200 elements.

 All backup files must now be specified in the correct order in 
 arrayOfFilesOrDevices.  Since all backup files are verified for integrity
 at the beginning of the restore, they all must be available when this method is
 called.

 When executed using Topaz, the result is either a String describing the success
 of the operation (in which case the Topaz result (obj **) may be nil), or an
 error message.

 This method terminates GemStone Smalltalk execution and does an automatic
 logout.  You will need to relogin to continue.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.

 A GciHardBreak during this method terminates the session.

 See restoreFromBackup: for further documentation."

^ self _restoreFrom: arrayOfFilesOrDevices opcode: 7
%


category: 'Backup and Restore'
method: Repository
restoreFromBackups: arrayOfFilesOrDevices newSystemUserPassword: aString

"Same as the restoreFromBackups: method except that a new password is
 set for SystemUser in the restored repository.  Passwords for all
 other user profiles will be restored to their values at the time 
 the backup was taken.

 If aString is nil, then this method functions the same as the
 restoreFromBackups: method and the SystemUser password is not
 changed.

 This method must be run as SysterUser if a new SysterUser password is
 specified."

^ self _restoreFrom: arrayOfFilesOrDevices opcode: 7 newSystemUserPassword: aString
%

category: 'Backup and Restore'
method: Repository
restoreFromCurrentLogs

"After a restoreFromBackup: returns true, this method may be executed to redo
 transactions which occurred since the backup was made.  This method re-does
 all transactions contained in the log files that are in Stone's current log
 directories or devices as defined by the STN_TRAN_LOG_DIRECTORIES and
 STN_REPL_TRAN_LOG_DIRECTORIES configuration file parameters.

 When executed using Topaz, the result is either a String describing the success
 of the operation (in which case the Topaz result (obj **) may be nil), or an
 error message.

 This method terminates GemStone Smalltalk execution and does an automatic
 logout.  You will need to relogin to continue.

 If some log files written since the restored backup file(s) were generated are
 no longer on-line, those off-line logs must be processed using the
 restoreFromLog: method before this method can be used.

 When opening a log file on a file system, if the filename ends in '.gz',
 the file is expected to be in gzip compressed format, and will be
 opened as such.  Restoring a compressed tranlog from a raw 
 partition is not supported.

 If GemStone was using partial-logging mode at the time restored backup
 file(s) were written then restoreFromCurrentLogs is not allowed.

 You must be the only user logged in, otherwise an error is generated.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.  This method puts the session into
 #manualBegin transaction mode.

 Note that restore status is an attribute of the Repository, not of a
 session, so the required preceding restoreFromBackup: could have been
 executed in some preceding session."

System transactionMode: #manualBegin .
System abortTransaction .

^ self _restoreFrom: nil opcode: 4 .
%
category: 'Backup and Restore'
method: Repository
restoreFromLog: fileOrDevice

"After a restoreFromBackup: returns true, this method may be
 executed to redo transactions which occurred since the backup was made.
 This method re-does all transactions contained in the specified log file.

 When executed using Topaz, the result is either a String describing the success
 of the operation (in which case the Topaz result (obj **) may be nil), or an
 error message.  
 
 This method terminates GemStone Smalltalk execution and does an automatic
 logout.  You will need to relogin to continue.

 Log files must be restored in time-sequence starting from the log file
 that was active at the time backup was made.

 When opening a log file on a file system, if the filename ends in '.gz',
 the file is expected to be in gzip compressed format, and will be
 opened as such.  Restoring a compressed tranlog from a raw 
 partition is not supported.

 Use the restoreStatus method to determine the next file required for a restore
 operation.  Note that restore status is an attribute of the Repository, not of
 a session, so the required preceding restoreFromBackup: could have been
 executed in some preceding session.

 If GemStone was using partial-logging mode at the time restored backup
 file(s) were written then restoreFromLog: is not allowed.

 You must be the only user logged in, otherwise an error is generated.

 This method requires the FileControl privilege.  It is recommended
 that you run as either DataCurator or SystemUser.  This method puts the
 session into #manualBegin transaction mode."

System transactionMode: #manualBegin .
System abortTransaction .

^ self _restoreFrom: fileOrDevice opcode: 5 .
%
category: 'Backup and Restore'
method: Repository
restoreNoReclaimFromLog: fileOrDevice

  "Same functionality as restoreFromLog:, howvever, no reclaim is performed
   during the restore operation. Use restoreReclaimPages to perform reclaims."

System transactionMode: #manualBegin .
System abortTransaction .

^ self _restoreFrom: fileOrDevice opcode: 10.
%
category: 'Backup and Restore'
method: Repository
restoreNoShadowsFromBackups: arrayOfFilesOrDevices

"Same as restoreFromBackups: except that partially filled data pages are
 not added to the list of scavengable pages upon completion of the restore.

 Using this method may cause the restored repository to perform faster than
 one restored using restoreFromBackups, especially immediately following
 the restore.

 See restoreFromBackups: for further documentation."


^ self _restoreFrom: arrayOfFilesOrDevices opcode: 12
%

category: 'Backup and Restore'
method: Repository
restoreNoShadowsFromBackups: arrayOfFilesOrDevices newSystemUserPassword: aString

"Same as the restoreNoShadowsFromBackups: method except that a new password is
 set for SystemUser in the restored repository.  Passwords for all
 other user profiles will be restored to their values at the time 
 the backup was taken.

 If aString is nil, then this method functions the same as the
 restoreNoShadowsFromBackups: method and the SystemUser password is not
 changed.

 This method must be run as SysterUser if a new SysterUser password is
 specified."

^ self _restoreFrom: arrayOfFilesOrDevices opcode: 12 newSystemUserPassword: aString
%

category: 'Backup and Restore'
method: Repository
restoreReclaimPages

  "Reclaim up to the number of pages specified by the stone
   configuration parameter STN_RECOVERY_PAGE_RECLAIM_LIMIT 
   (default 2000).  To be used in conjunction with restoreNoReclaimFromLog: 
   to help ensure predictable restore times.  Returns number of pages 
   actually reclaimed."

^ self _restoreFrom: ' ' opcode: 11.
%
category: 'Backup and Restore'
method: Repository
restoreStatus

"Returns a String describing the current restore status of the Repository,
 including the next transaction log file or backup file required to continue
 the restore.  

 Restore status is an attribute of the Repository, not of the session, 
 and persists across logout/login and stopstone/startstone.

 This method requires the FileControl privilege."

^ self _restoreStatusInfo at: 1
%
category: 'Backup and Restore'
method: Repository
restoreStatusNextFileId

"Returns a SmallInteger, the fileId of the next transaction log or backup that
 should be restored, or nil if restore not active.

 Restore status is an attribute of the Repository, not of the session, 
 and persists across logout/login and stopstone/startstone.

 This method requires the FileControl privilege."

^ self _restoreStatusInfo at: 3 .
%
category: 'Backup and Restore'
method: Repository
setArchiveLogDirectories: arrayOfDirectorySpecs
tranlogPrefix: tranlogPrefixString
replicateDirectories: arrayOfReplicateDirSpecs
replicatePrefix:  replicPrefixString

"Specifies the directories and raw partitions to be searched by subsequent
 invocation(s) of restoreFromArchiveLogs.

 The argument arrayOfDirectorySpecs must be an Array of one or more Strings.
 Each String must name a file system directory or raw device specification.
 An error is generated if any of the directories or devices does not exist.
 It is not an error if they exist but do not yet contain any transaction logs.

 The argument tranlogPrefixString must be a String, the file prefix to be used 
 when searching for transaction logs in file systems specified in
 arrayOfDirectorySpecs.  The argument may be nil, in which case the value for
 STN_TRAN_LOG_PREFIX in the Stone's configuration file is used.

 The argument arrayOfReplicateDirSpecs must be nil, or an Array of zero or more
 Strings that specify file system directories and/or raw devices to search for
 replicate transaction logs that have the file prefix replicPrefixString.  An
 error is generated if any of the directories or devices does not exist.  It is
 not an error if they exist but do not yet contain any transaction logs.

 The argument replicPrefixString must be a String, the file prefix to be used 
 when searching for transaction logs in file systems specified in
 arrayOfReplicateDirSpecs.  The argument may be nil, in which case the value for
 STN_REPL_TRAN_LOG_PREFIX in the Stone's configuration file is used.

 Does not require privileges.  Does not require that you be the only user
 logged in.  However, a subsequent restore operation to use the state set by
 this method will require that the Repository be in restore state, and that you
 have FileControl privilege and that you then be the only user logged in."
 
| validateStringBlk |
validateStringBlk := [:aString | 
  (aString class isSubclassOf: String) ifFalse:[ 
    aString _errorExpectedClass: String 
    ].
  aString size > 1023 ifTrue:[ aString _error: #rtErrArgOutOfRange ].
  ].
arrayOfDirectorySpecs _validateClass: Array .
arrayOfDirectorySpecs size < 1 ifTrue:[
  arrayOfDirectorySpecs _error: #errArgTooSmall args:#[ 1 ].
  ].
arrayOfDirectorySpecs do:[ :aString | validateStringBlk value: aString  ].
tranlogPrefixString ~~ nil ifTrue:[
  validateStringBlk value: tranlogPrefixString .
  ].
arrayOfReplicateDirSpecs ~~ nil ifTrue:[
  arrayOfReplicateDirSpecs _validateClass: Array .
  arrayOfReplicateDirSpecs do:[ :aString | validateStringBlk value: aString  ].
  replicPrefixString ~~ nil ifTrue:[
    validateStringBlk value: replicPrefixString .
    ].
  ].

^ self _primSetArchiveLogDirectories: arrayOfDirectorySpecs
        tranlogPrefix: tranlogPrefixString
        replicateDirectories: arrayOfReplicateDirSpecs
        replicatePrefix:  replicPrefixString
%
category: 'Updating'
method: Repository
size: anInteger

"Changes the indexed size of the receiver to anInteger.  May not be used to
 grow the receiver.  Growing a Repository should only be done by sending the
 following message:

 Segment newInRepository: SystemRepository

 If anInteger is less than the current size of the receiver, the receiver is
 shrunk accordingly.  If anInteger is greater than the current size of the
 receiver, an error is generated."

anInteger > self size
  ifTrue:[ ^ self _errorIndexOutOfRange: anInteger ].

^ super size: anInteger
%
category: 'Transaction Logging'
method: Repository
_startNewLog: forceBoolean

"Causes the most current transaction log to be closed and a new transaction log
 to be opened for writing.  The location of the new log is controlled by the
 STN_TRAN_LOG_DIRECTORIES configuration file parameter.

 If GemStone is running in partial logging mode, then a preceding transaction
 log may be deleted.  See documentation on the STN_TRAN_FULL_LOGGING
 configuration parameter for more details.

 If forceBoolean is false and a checkpoint is in progress the operation,
 the NOT performed and a -1 is returned.  

 If forceBoolean is true, the session will block while the stone immediately
 completes any checkpoint in progress and a new log is created.

 Returns the fileId of the new log or -1 if a new log could not be started.

 This method requires the FileControl privilege."

<primitive: 432>

self _primitiveFailed: #_startNewLog:.
self _uncontinuableError
%

category: 'Transaction Logging'
method: Repository
startNewLog

"Causes the most current transaction log to be closed and a new transaction log
 to be opened for writing.  The location of the new log is controlled by the
 STN_TRAN_LOG_DIRECTORIES configuration file parameter.

 If GemStone is running in partial logging mode, then a preceding transaction
 log may be deleted.  See documentation on the STN_TRAN_FULL_LOGGING
 configuration parameter for more details.

 If a checkpoint is in progress the operation is NOT performed and a -1
 is returned.  If it is successful the fileId of the new log is returned.

 This method requires the FileControl privilege."

^ self _startNewLog: false
%

category: 'Transaction Logging'
method: Repository
forceStartNewLog

"Causes the most current transaction log to be closed and a new transaction log
 to be opened for writing.  The location of the new log is controlled by the
 STN_TRAN_LOG_DIRECTORIES configuration file parameter.

 If GemStone is running in partial logging mode, then a preceding transaction
 log may be deleted.  See documentation on the STN_TRAN_FULL_LOGGING
 configuration parameter for more details.

 If a checkpoint is in progress, the session will block while the stone immediately
 completes the checkpoint the new log is created.

 Returns the file ID of the new transaction log.

 This method requires the FileControl privilege."

^ self _startNewLog: true
%

category: 'Backup and Restore'
method: Repository
timeToRestoreTo: aDateTime

"Sets the time at which restoreFromLog: and restoreFromCurrentLogs will stop.
 The restore will stop at the first checkpoint which originally occurred at or
 after aDateTime.  If timeToRestoreTo: has not been used since
 restoreFromBackup: completed, then restores will proceed to the end of the
 specified transaction log(s).

 An error is generated if aDateTime precedes the time of the last restored
 checkpoint, as shown by restoreStatus.  An error is generated if the receiver
 is not in restore-from-log state.

 Execution of restoreFromBackup: or commitRestore will cancel the effect of any
 previous execution of timeToRestoreTo:.

 If restore has stopped at a time specified by this method, then a subsequent
 restoreFromLog: or restoreFromCurrentLogs may be used to continue restoring
 past the time specified by the last timeToRestoreTo:.  Alternatively,
 timeToRestoreTo: can be used to specify another point in time before 
 continuing the restore.

 This method requires the FileControl privilege."

| aTimeT |

"Convert offset from 1901 to 1970 using constant derived from
 (DateTime newGmtWithYear: 1970 dayOfYear: 1 milliseconds: 0) asSecondsGmt"

aTimeT := aDateTime asSecondsGmt - 2177452800.

(aTimeT < 0 _or:[ aTimeT > 16r7fffffff]) ifTrue:[
  aDateTime _error: #rtErrArgOutOfRange 
  ].

^ self _restoreFrom: aTimeT opcode: 8
%
category: 'Repository Conversion'
method: Repository
transformAllObsSym: obsSymSet

"Private. Attempts to transform all the ObsoleteSymbols contained in obsSymSet
 to Symbols without changing their identities.

 Returns a three-element Array whose contents are as follows:

 (1) The first element contains a list of all the ObsoleteSymbols that could
 be successfully transformed. 

 (2) The second element contains an Array of all the ObsoleteSymbols that
 could not be transformed. 

 (3) The third element contains an Array of Symbols whose contents are the
 corresponding ObsoleteSymbol contained in the Array returned in (2) above.
 
 The Arrays in second and third elements can be used for replacing the 
 ObsoleteSymbols with Symbols."

| convertedArray oldSymArray newSymArray |

convertedArray := Array new.
oldSymArray := Array new.
newSymArray := Array new.

obsSymSet do: [ :oldSym | | newSym |
  newSym := oldSym transformIntoSymbol.
  (newSym == oldSym) 
    ifTrue: [ 
      "Transformation was successful."
      convertedArray add: newSym 
      ]
    ifFalse: [
      oldSymArray add: oldSym.
      newSymArray add: newSym.
      ]
  ].

^ #[ convertedArray, oldSymArray, newSymArray ].
%
category: 'Extent Operations'
method: Repository
validateExtentId: anExtentId

"Returns the argument if it is valid, otherwise generates an error."

| maxExtentId |
anExtentId == nil ifTrue:[ ^ anExtentId ] .
anExtentId _validateClass: SmallInteger.
(anExtentId < 1) ifTrue:
   [anExtentId _error: #rtErrArgOutOfRange .  ^ 1 ].
maxExtentId := self numberOfExtents .
(anExtentId > maxExtentId) ifTrue:
   [anExtentId _error: #rtErrArgOutOfRange .  ^ maxExtentId ].
^ anExtentId .
%
category: 'Storing and Loading'
method: Repository
writeTo: aPassiveObject

"Instances of Repository cannot be converted to passive form.  This method
 writes nil to aPassiveObject and stops GemStone Smalltalk execution with a
 notifier."

aPassiveObject writeObject: nil.
self _error: #rtErrAttemptToPassivateInvalidObject
%
category: 'Private-Repository Analysis'
method: Repository
_buildLimitSetForRefPathScan

"Build a list of objects to serve as the limit set for a reference path scan.
 The reference path scan will end when a reference from any object returned
 by this method is found during the scan."

|set|

set := IdentitySet new.
set addAll: SystemRepository. "Add all segments"
"Get all symbol list dictionaries and all versions of all classes
 from all user profiles.  Use a set prevents duplicates."
1 to: AllUsers size do:[:m| |symbolList|
  symbolList := (AllUsers at: m) symbolList.
    1 to: symbolList size do:[:n| |eachSymbolListDictionary|
      eachSymbolListDictionary := symbolList at: n.
      (set includes: eachSymbolListDictionary)
         ifFalse:[
            set add: eachSymbolListDictionary.
            eachSymbolListDictionary keysAndValuesDo:[:k :v|
              v isBehavior
                ifTrue:[
                  1 to: v classHistory size do:[:p| |class|
                    class := v classHistory at: p.
                    set add: class.
                    set add: class class.].
                ].
            ].
         ].
    ].
].
set add: (Globals at: #SharedDependencyLists).
set add: (Globals at: #AllSymbols).
^set asArray
%
category: 'Private'
method: Repository
_buildWeakRefSet

"Build the hiddenSet containing the objects that contain weak references.  Each
 object referenced in the global GcWeakReferences is sent the message so that it
 can add its components that contain weak references to the hiddenSet."

System _hiddenSetReinit: 35.
GcWeakReferences do: [:each | each _buildWeakRefSet].
^self
%
category: 'Private'
method: Repository
_contentsOfPage: aPage
 
"If aPage is an invalid page ID, then the primitive fails.  Returns an Array of
 size at least 5 for all kinds of pages.  For data pages, the Array size is at
 least 15 and varies depending upon the number of objects in the page.

 The aPage argument should be a positive Integer specifying a valid page ID.
  
 The contents of the returned Array are as follows:

    index       description
     1.         pageKind
     2.         sessionId for user that wrote the page.
     3.         low beginId  (transaction begin ID)
     4.         high beginId
     5.         bytesUsed
                     (elements 6 and above are only present for data pages )
     6.         approximate page creation time (a DateTime)
     7.         clusterId
     8.         freeBytes
     9.         numFreeChunks  
    10.         sizeLargestFree
    11.         numUsedBytes   
    12.         numberOfObjects
    13.         maxObjSize
    14.         minObjSize
    15.         numberOfValidObjects  
    16.         numberOfInvalidObjects
    (17.. 17 + numberOfValidObjects - 1) list of valid objects  
    (17 + numberOfValidObjects ...)  list of OOPs of invalid objects.  "

<primitive: 399>
self _primitiveFailed: #_pageContents:
%
category: 'Transaction Logging'
method: Repository
_currentLogInfo

"Returns the result specified for Repository | _logInfo: for the
 currently active transaction log."

| numLogs logInfo |

numLogs := (System configurationAt:#STN_TRAN_LOG_DIRECTORIES) size .
2 timesRepeat:[ "cope with changing to the next log"
  1 to: numLogs do:[ :j |
    logInfo := self _logInfo: j .
    (logInfo at: 3) == true ifTrue:[
      ^ logInfo 
      ].
    ].
  ].
self _halt:'Unable to determine the current log'.
^ #[ ]
%
category: 'Repository Usage Reporting'
method: Repository
_dbfStatistics: extentFilename

"Returns an Array of size 2 containing the number of bytes the given extent
 requires for physical size and the number of bytes of free space in the
 extent."

<primitive: 283>

extentFilename _validateClass: String .
^ self _primitiveFailed: #_dbfStatistics:
%
category: 'Garbage Collection'
method: Repository
_deadNotReclaimedCount

"Return the number of dead objects pending reclamation."

<primitive: 528>
self _primitiveFailed: #_deadNotReclaimedCount
%
category: 'Backup and Restore'
method: Repository
_destroyOtAndRestoreFrom: fileOrDevice

"Private.  Use this method at your own peril.

 Delete all existing objects and then do a full restore from the specified file.
 You must be the only user logged in and you must be SystemUser.  Logins are
 suspended during the execution of this method.  This commits the restore
 automatically and kills your session.

 This is used to create the smallest possible resulting Repository from a
 restore, and is intended for use during filein of kernel classes or of a
 schema, such as to generate a master Repository for distributing with a
 packaged application.

 This method requires the #FileControl privilege."

^ self _restoreFrom: fileOrDevice opcode: 3
%
category: 'Repository Usage Reporting'
method: Repository
_extentStatistics: extentIndex

"Returns an Array of size 2 containing the number of bytes the
 given extent requires for physical size and the number of bytes of
 free space in the extent."

<primitive: 286>

(extentIndex _isSmallInteger)
  ifTrue:[ ^ self _errorIndexOutOfRange: extentIndex]
  ifFalse:[ extentIndex _validateClass: SmallInteger ].
^ self _primitiveFailed: #_extentStatistics:
%
category: 'Private'
method: Repository
_findAllIndexObjs
"Scans the entire repository for index objects and returns an array containing
 2 arrays.  The first array contains all collections in the repository with 1
 or more indices.  The second array contains all objects with
 a DependencyList.  Dead objects are excluded from both result arrays.

 Garbage collection gems should be shutdown while this method is in progress."
<primitive: 530>
^ self _primitiveFailed: #_findAllIndexObjs	
%
category: 'Private'
method: Repository
_findConnectedObjs
"Return an array of all objects reachable starting from the objects
 in hidden set 26.  Objects implicitly referenced from the
 header such as the object's class and segment are not included in the
 scan.

 This method requires the GarbageCollection privilege."

<primitive: 521>
^ self _primitiveFailed: #_findConnectedObjs	
%
category: 'Private-Repository Analysis'
method: Repository
_findReferencePathToObject: anObject

"Finds a reference path to a single object.  See the comments in method
 Repository>>_findReferencePathToObjs: limitObjArray:  findAllRefs: printToLog:
 for more information.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."


|result limitArray|

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

limitArray := self _buildLimitSetForRefPathScan.
result := self _findReferencePathToObjs: (Array with: anObject)
               limitObjArray: limitArray
               findAllRefs: true
               printToLog: true.
result == nil
  ifTrue:[^nil].
^result first.
%
category: 'Private-Repository Analysis'
method: Repository
_findReferencePathToObjects: anArray findAllRefs: findAllRefs printToLog: printToLog

"Reference path analysis is an experimental new feature and is not
 intended for mission-critical use.

 Finds a reference path to an array of objects.  Returns an array
 of arrays showing how each object is connected to the repository
 by reference.  The first element of a reference path will usually
 be (see below) one of the objects in the limit set, which is computed
 by the method  Repository>>_buildLimitSetForRefPathScan.  The last
 object in a reference path will be anObject.

 Only a single reference path to each object is returned. It's possible
 for any object to be connected to the repository through multiple
 reference paths.

 References from objects in the stone's list of deadNotReclaimed objects
 are excluded from the scan and it is an error if any input object is in
 this list.  In this case the method will return nil.

 Only references by committed objects are checked.  It is an error if any
 input object is not committed.

 It is possible for an object to be disonnected from the repository and
 therefore have no reference path.  Disconnected objects are eligible for
 removal after the next markForCollection operation.  If an input object
 is found to be disconnected from the repository, the second element of
 the result array will be false, indicating no reference path was found.

 If findAllRefs is false, the scan will return when the first complete
 reference path is found for any input object.   If findAllRefs is true,
 the scan will continue until all reference paths have been analyzed.

 If printToLog is true, messages showing progress of the scan will be
 printed to stdout of the gem calling this method.  One message will
 be printed after each level of the reference path is analyzed for
 each object.

 The progressCount cache statistic for the gem calling this method will
 indicate how many objects have been processed for each scan of the 
 repository and is reset to zero at the end of each scan.

 This method operates entirely inside a transaction and will start a
 transaction if necessary.  This method may run for long periods of time
 on large repositories and should not be used on production systems.

 This method should be run as SystemUser to avoid errors due to
 object level security.

 A reference path scan for an object will be terminated and considered
 completed if any class or segment object is found in the reference path,
 even if the class or segment is not present in the limit set.  Normally
 all classes and segments are connected to the repository and are
 automatically be included in the limit set.  However, for garbage collection
 purposes, disconnected classes and segments are considered live objects if
 any other object references them.  In other words, a class object is a live
 object if any live instances of that class exist in the repository. 
 Similarly, a segment is live if any object in the repository resides in that
 segment.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 This method returns an array of arrays.  Each elemet of the returned array
 corresponds to an object in the input array with the same index and has the
 following format:
 1 - the search object
 2 - aBoolean - true if the reference path was found, false if not.
  (if element 2 is true, the remaining elements comprise the reference path)
 3 - first element of the reference path (an object in the limit set)
 4 - second element of the reference path.
 ...
 n - last element of the reference path, which will be the search object."

| limitArray |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

limitArray := self _buildLimitSetForRefPathScan.
^self _findReferencePathToObjs: anArray
 limitObjArray: limitArray
 findAllRefs: findAllRefs
 printToLog: printToLog
%
category: 'Private-Repository Analysis'
method: Repository
_findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: aBoolean printToLog: anotherBoolean

<primitive: 548>
anArray _validateClass: Array.
limitArray _validateClass: Array.
aBoolean _validateClass: Boolean
anotherBoolean _validateClass: Boolean.
(anArray size ==  0)
  ifTrue:[ ^anArray  _error: #rtErrArgOutOfRange  ] .
(limitArray size ==  0)
  ifTrue:[ ^limitArray _error: #rtErrArgOutOfRange  ] .
^self _primitiveFailed: #_findReferencePathToObjs:limitObjArray:findAllRefs:printToLog: .
%
category: 'Private'
method: Repository
_fullBackupTo: fileOrDevice
  backupId: backupIdOrNil
  MBytes: mByteLimit
  firstFile: firstFileBool
  compressed: compressBool

"Private.  Provides implementation of backup methods.

 Returns true if the backup was completed.  Returns a message (a String) if
 continueFullBackupTo:MBytes: should be run to complete the backup.

 A GciHardBreak during this method will terminate the session."

  <primitive: 394>

  fileOrDevice _validateClass: String.
  backupIdOrNil _validateClass: SmallInteger.  "should be positive"
  mByteLimit _validateClass: SmallInteger.
  firstFileBool _validateClass: Boolean .
  compressBool _validateClass: Boolean .
  self _primitiveFailed: #_fullBackupTo:backupId:MBytes:firstFile:compressed:
%
category: 'Repository Conversion'
method: Repository
_getContainerList: obsSymSet result: anArray

"Private. Returns a list of all the objects that contain one or more of the
 ObsoleteSymbols contained in obsSymSet."

<primitive: 290>

self _primitiveFailed: #_getContainerList: .
self _uncontinuableError
%
category: 'Private'
method: Repository
_listInstances: anIdentitySet

"Private.  Applications should use the Repository>>listInstances: method."

"Returns a list of instances on the receiver that belong to one of the classes
 listed in anArray.  The result of this method is an Array of IdentitySets, where the
 contents of each set consists of all instances whose class is equal to the
 corresponding element in anArray.

 Warning: you may retrieve instances to which you have no read access, so this
 method is mostly of use to SystemUser.

 The argument aSet must not be larger than 2024 elements.

 Scans the entire Repository."

<primitive: 392>

anIdentitySet _validateClass: IdentitySet.
(anIdentitySet size > 2024 "virtual machine constant")
  ifTrue:[ anIdentitySet _error: #rtErrArgOutOfRange  ] .
anIdentitySet do: [:each | each _validateClass: Behavior].
^self _primitiveFailed: #listInstances:
%
category: 'Private'
method: Repository
_listObjectsInSegments: anIdentitySet

"Private.  Applications should use the Repository>>listObjectsInSegments: method."

"Returns a list of committed objects that belong to one of the segments listed
 in anArray.  The result of this method is an Array of IdentitySets, where 
 the contents of each set consists of all instances whose Segment is 
 the corresponding element in anArray.  Dead objects (those in the stone's
 deadNotReclaimed set) are not included in the scan results.

 Warning: you may retrieve instances to which you have no read access, so this
 method is mostly of use to SystemUser.

 The argument aSet must not be larger than 2024 elements.

 Scans the entire Repository once."

<primitive: 569>

anIdentitySet _validateClass: IdentitySet.
(anIdentitySet size > 2024 "virtual machine constant")
  ifTrue:[ anIdentitySet _error: #rtErrArgOutOfRange  ] .
anIdentitySet do: [:each | each _validateClass: Segment].
^self _primitiveFailed: #_listObjectsInSegments:
%
category: 'Class Management'
method: Repository
_listReferences: anArray withLimit: aSmallInt

"Private. Returns a list of instances in the receiver that have a reference to 
 one of the objects specified in anArray.  The result of this method is an 
 Array of Arrays, where the contents of each Array consists of instances that 
 have a reference to the corresponding element in anArray.  The number in each 
 Array is limited by aSmallInt.  If an instance for which you have no read
 authorization is found, the first slot in the Array contains 'true'; otherwise
 the first slot contains 'false'.

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 This method is very expensive.  It scans the entire Repository and looks
 at every instance variable of every object."

<primitive: 393>
aSmallInt _validateClass: SmallInteger.
anArray _validateClass: Array.
(anArray size > 2024 "virtual machine constant")
  ifTrue:[ anArray _error: #rtErrArgOutOfRange  ] .
^self _primitiveFailed: #listReferences:
%
category: 'Transaction Logging'
method: Repository
_logInfo: logDirId

"If logDirId is within the range of 1 to number of log directories, returns an
 Array containing the following instance variables.  Otherwise returns nil.

 1: logDirId (SmallInteger)
 2: numLogDirs (SmallInteger)
 3: isActiveLog (Boolean)
 4: fileId (SmallInteger) (if active log )
 5: fileSize (SmallInteger) in units of MBytes (if active log)
 6: maxFileSize (SmallInteger) in units of MBytes
 7: logOriginTime (Integer)  (if active log)
 8: fileName (String) (fileOrDevice if active log, else directoryOrDevice)
 9: replicateName (String) (fileOrDevice if active log, else directoryOrDevice)
10: fileId of begin record of oldest transaction in current checkpoint,
         if active log, otherwise nil"

<primitive: 397>

logDirId _validateClass: SmallInteger.
^ self _primitiveFailed: #_logInfo:
%
category: 'Garbage Collection'
method: Repository
_markGcCandidates

"Primitive to perform the analysis."

<primitive: 324>
^ self _primitiveFailed: #_markGcCandidates
%
category: 'Repository Conversion'
method: Repository
_migrateGroups

"Migrates group collections in all the Segments of the receiver."

self do:[ :aSegment | aSegment _migrateGroups ].
%
category: 'Garbage Collection'
method: Repository
_objectAuditWithLimit: statsLimit fullChecks: aBoolean

  "Similar to Repository | objectRepairWithLimit:destroyIndexes:,
   except that any errors found are not repaired.
   A GciHardBreak during this method will terminate the session."

<primitive: 389>
^ self _primitiveFailed: #_objectAuditWithLimit:fullChecks:
%
category: 'Garbage Collection'
method: Repository
_objectRepairWithLimit: statsLimit destroyIndexes: aBoolean fullChecks: aBoolean2

"Performs an object audit on all objects in GemStone.  This audit is to be
 done after the garbage collection, shrink and page audit have completed
 successfully.

 The following checks are done on each object:

 1.  Object reference consistency - no object should contain any reference to a
     non-existent object.
 2.  All OOPs not found in the object table should be in the freeOopList.
 3.  The physical implementation of each object is consistent with its class.

 The following repairs are done:

 1.  Free OOPs found in the Object Table are removed and inserted into the
     FreeOopList (if they are not already there).
 2.  Store OopClassString for class field of byte Objects, OopClassArray for
     pointer objects, and OopClassBag for NSC objects.  If the object has a
     dependency tag, store OopNil in the tag to dereference the dependency
     list.

 If aBoolean is true, all indexes in GemStone are removed.

 If the fullChecks boolean is true, the method will return with an error if
 not in single-user mode or if reclaimAll cannot be completed.  Refer to
 comment in repairWithLimit:fullChecks: for details.

 All errors found and any repairs done are reported in the file AUDIT_LOG in
 the logical directory GEMSTONE.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 A GciHardBreak during this method will terminate the session."

<primitive: 388>
^ self _primitiveFailed: #_objectRepairWithLimit:destroyIndexes:fullChecks:
%
category: 'Garbage Collection'
method: Repository
_pagesWithPercentFree: aPercent doScavenge: aBoolean

"This method returns an Array containing the following statistics:

 1.  The total number of data pages processed.
 2.  The sum in bytes of unused space in all data pages.  This quantity is a
     measure of data fragmentation in the receiver.
 3.  The number of bytes in a page.
 4.  The number of data pages that have at least the specified percentage of
     unused space.

 If aBoolean is true, then the primitive schedules for scavenging the pages
 that have at least the specified percentage of unused space.

 Do not confuse unused space on a page with free space (unused pages) in a
 Repository or Extent.  See the 'freeSpace' and 'freeSpaceInExtent:' methods
 for more information.

 In order to execute this method you must have the GarbageCollection
 privilege.

 A GciHardBreak during this method terminates the session."

<primitive: 390>
^ self _primitiveFailed: #_pagesWithPercentFree:doScavenge
%
category: 'Garbage Collection'
method: Repository
_primMarkForCollection: waitTimeSeconds

<primitive: 431>

  self _primitiveFailed: #_primMarkForCollection: .
  self _uncontinuableError
%
category: 'Private'
method: Repository
_primSetArchiveLogDirectories: arrayOfDirectorySpecs
tranlogPrefix: tranlogPrefixString
replicateDirectories: arrayOfReplicateDirSpecs
replicatePrefix:  replicPrefixString

"Private."

"Send archive log directories specification to Stone.  Sender has
 done class kind checks of all the arguments.  Stone will check for
 existence of each directory.  arrayOfDirectorySpecs should be non-empty.
 The other arguments may be nil.  If tranlogPrefixString or replicPrefixString
 are nil, the applicable configuration file parameter is substituted."

<primitive: 469>
self _primitiveFailed: #_primSetArchiveLogDirectories:
%
category: 'Garbage Collection'
method: Repository
_reinitializeGcCandidates

"Reset the GcCandidates objects.  Must be run as SystemUser"

| oldQueue oldCount newQueue newCount newSize |

oldQueue := Globals at: #GcCandidates.
oldCount := Globals at: #GcCandidatesCount.
newSize := oldQueue maxSessionId max: oldCount maxSessionId.
newQueue := (oldQueue class new: newSize).
newCount := (oldCount class new: newSize).
newQueue changeToSegment: ( oldQueue segment ).
newCount changeToSegment: ( oldCount segment ).
Globals at: #GcCandidates put: newQueue.
Globals at: #GcCandidatesCount put: newCount.
%
category: 'Repository Conversion'
method: Repository
_replaceAllObsSym: obsSymSet

"Private. Replaces all the ObsoleteSymbols listed in obsSymSet in all
 the objects in the System with their Symbol counterparts."

<primitive: 290>

self _primitiveFailed: #_replaceAllObsSym: .
self _uncontinuableError
%
category: 'Repository Conversion'
method: Repository
_replaceAllSymIn: containerList obsSymList: oldSymList mappingDict: mappingDict

"Private. Replaces ObsoleteSymbols in all objects contained in containerList.
 (containerList is an Array of objects which contain ObsoleteSymbols. obsSymList is anArray of ObsoleteSymbols to be replaced. mappingDict is a dictionary which
 contains the corresponding (new) Symbols."

<primitive: 486>

self _primitiveFailed: #_replaceAllSymIn:obsSymList:mappingDict: .
self _uncontinuableError
%

category: 'Backup and Restore'
method: Repository
_restoreFrom: fileOrDevice opcode: anInt
^ self _restoreFrom: fileOrDevice opcode: anInt newSystemUserPassword: nil
%

category: 'Backup and Restore'
method: Repository
_restoreFrom: fileOrDevice opcode: anInt newSystemUserPassword: aString

"Private.  GemStone internal use only.  Use this method at your own peril.

 opcodes: 0 = restoreFromBackup.  Shuts down the GcGem.
          1 = abortRestore
          2 = commitRestore
          3 = destroy existing objects prior to restore; this requires that you
              be SystemUser, that you be the only user, and that the backup
              being restored be in a single file.  This commits the restore
              automatically and kills your session, and is used to create the
              smallest possible resulting Repository.  Shuts down the GcGem.
          4 = restore from current logs.
          5 = restore from specified log device or directory.
          6 = _restoreStatusInfo 
          7 = restore multiple backup files.  fileOrDevice is an Array of
              files or devices.
          8 = timeToRestoreTo: aGmtTime (SmallInteger). 
          9 = restore from archive logs.
         10 = restore from specified log device or directory with no reclaim.
         11 = reclaim a chunk of pages (2000).
         12 = restore multiple backup files. Does not add partially filled pages
              to the scavengable pages list.

 This method must be run as SysterUser if a new SysterUser password is
 specified."

<primitive: 395>
| expectedClass argSize |
anInt _validateClass: SmallInteger.
((anInt > 12) _or:[anInt < 0] )
  ifTrue:[anInt _error: #rtErrArgOutOfRange "bad opcode"].

anInt == 8
  ifTrue:[expectedClass := Integer]
  ifFalse:[
    (#[1,2,4,6,9] includesIdentical: anInt)
      ifTrue:[expectedClass := UndefinedObject]
      ifFalse:[
        (#[0,3,5,10,11] includesIdentical: anInt)
           ifTrue:[expectedClass := String]
           ifFalse:[
               (#[7,12] includesIdentical: anInt)
                   ifTrue:[expectedClass := Array].
           ].
      ].
].
fileOrDevice _validateClass: expectedClass.

expectedClass == Integer
  ifTrue:[
(fileOrDevice < 0 _or:[ fileOrDevice > 16r7fffffff])
   ifTrue:[ fileOrDevice _error: #rtErrArgOutOfRange ].
].

expectedClass == String
  ifTrue:[
    argSize := fileOrDevice size.
    (argSize < 1 _or:[ argSize > 1000]) 
         ifTrue:[ fileOrDevice _error: #rtErrArgOutOfRange ] .
].

expectedClass == Array
  ifTrue:[
    argSize := fileOrDevice size.
    argSize <= 0  ifTrue:[ fileOrDevice _error: #objErrCollectionEmpty ] .
    argSize > 200 ifTrue:[ fileOrDevice _error: #rtErrArgOutOfRange ].
    1 to: argSize do:[:n| |aFileSpec size|
       aFileSpec := argSize at: n.
       aFileSpec _validateClass: String .
       size := aFileSpec size .
       (size < 1 _or:[size > 1000]) 
           ifTrue:[aFileSpec _error: #rtErrArgOutOfRange ] .
  ] .
].

aString == nil
  ifFalse:[ aString _validateClass: String ].

self _primitiveFailed: #_restoreFrom:opcode:newSystemUserPassword: .
self _uncontinuableError
%

category: 'Backup and Restore'
method: Repository
_restoreStatusInfo

"Returns an Array describing the current restore status of the Repository,
 including the next transaction log file or backup file required to continue
 the restore.  The Array contains the following 4 elements:

   1: a String describing the restore status.
   2: a SmallInteger, 1 = active from backup, 2 = active from log
   3: a SmallInteger, next fileId to restore 
   4: a SmallInteger, next blockId to restore

 If restore is not active, the elements 2 to 4 are nil.

 Restore status is an attribute of the Repository, not of the session, 
 and persists across logout/login and stopstone/startstone.

 This method requires the FileControl privilege."

^ self _restoreFrom: nil opcode: 6 .
%
category: 'Garbage Collection'
method: Repository
_scanForParentObjects

"Scan the repository to find all objects that reference any object
 contained in hidden set 26.  Place the resultant objects in 
 hidden set 27.  Dead objects are not included in the results.

 This method may only be run by SystemUser.

 Refer to the methods in class System for information about accessing
 the hidden sets.

 Returns true if successful, false if hidden set 26 is empty."

<primitive: 529>
self _primitiveFailed: #_scanForParentObjects .
%
category: 'Updating'
method: Repository
_setVersion

"This method writes the Stone executable version string to the receiver.

 If the Repository was in conversion from 4.1 to 5.0, this clears
 the Repository conversion flag in the Repository root page and re-enables
 automatic startup of the GcGem."

<primitive: 289>

self _primitiveFailed: #_setVersion .
self _uncontinuableError
%
category: 'Garbage Collection'
method: Repository
_shadowPagesByExtent

"Returns an array of integers where each element represents the number of
 shadow pages in that extent.  The first element is the number of
 shadow pages in extent 0, the second the number of shadow pages in extent 1,
 etc."
<primitive: 526>
self _primitiveFailed: #_shadowPagesByExtent
%
category: 'Garbage Collection'
method: Repository
_shadowPagesCount

"Returns the approximate number of shadow pages present in the system."
<primitive: 527>
self _primitiveFailed: #_shadowPagesCount
%
category: 'Private'
method: Repository
_shrinkExtents

"Private. This method should be used with caution - errors occurring during
 operation of the method cannot be recovered from and may require restoring
 from backup.
 
 Truncate all Extents of the Repository to remove internal free space between
 the last used page in each extent and the end of the file containing the
 extent.  Has no effect for extents on a raw disk partition.  Note that it 
 does not recover internal free space occurring between pages in use.

 This may be run while the system is in normal operation and does not commit
 the current transaction.  System performance may be degraded while this
 operation is in progress.

 This method requires the FileControl privilege.  It is recommended that you
 run as either DataCurator or SystemUser.

 If DBF_PRE_GROW is enabled in the configuration file then this the extents
 will be grown again the next time Stone is restarted, thus cancelling the
 effect of this method.

 The Garbage Collector session is shut down for the duration of this method."

<primitive: 427>

self _primitiveFailed: #_shrinkExtents.
self _uncontinuableError
%

category: 'Repository Conversion'
method: Repository
_writeGs64UpgradeDataToDirectory: aString
"Create various files required for conversion.  aString must be a
 an instance of String which represents the path to a directory
 which is writable by this session."

<primitive: 553>
self _primitiveFailed: #_writeGs64UpgradeDataToDirectory: .
self _uncontinuableError.
%


category: 'Cache Preloading'
method: Repository
readObjectTableForGem: gemNum of: totalGems
"This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 See the comments in method
   Repository>>_readObjectTableFromOopNum: anInt toOopNum: aBiggerInt            
 for more information."

|hwMark delta start end|

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

hwMark := System _oopNumberHighWaterMark - 1 .

delta := hwMark // totalGems .
start := (delta * (gemNum - 1)) + 1 .

gemNum == totalGems
	ifTrue:[end := hwMark]
	ifFalse: [end := gemNum * delta] .

^self _readObjectTableFromOopNum: start toOopNum: end .
%


category: 'Cache Preloading'
method: Repository
_readObjectTableFromOopNum: anInt toOopNum: aBiggerInt

"Read a section of the object table from anInt up to aBiggerInt -1.  Both 
 arguments must either be a SmallInteger or LargePositiveInteger.

 If the shared page cache becomes full before all pages have been read, this method
 will return false and print a warning message to the log file.  The shared cache is
 determined to be full if the number of free frames in the cache falls below
 the target free frame limit + 20%.  The target free frame limit is determined
 by the SHR_TARGET_FREE_FRAME_COUNT configuration file parameter.

 The session will run outside of transaction and will respond to SigAborts.

 Returns true if all object tables were loaded.  Returns false if the loaded was
 terminated early because the shared page cache is low on free frames."

<primitive: 568>
(anInt _isSmallInteger)
  ifFalse:[anInt _validateClass: LargePositiveInteger].
(aBiggerInt _isSmallInteger)
  ifFalse:[aBiggerInt _validateClass: LargePositiveInteger].
(anInt < aBiggerInt)
  ifFalse:[anInt _error: #rtErrArgOutOfRange].
^self _primitiveFailed: #_readObjectTableFromOopNum:toOopNum: .
%

category: 'Class Management'
method: Repository
countInstances: anArray

"Returns a count of instances on the receiver that belong to one of the
 classes listed in anArray.  The result of this method is an Array of Associations,
 where each key is an element of the input array and the value is the count of all 
 instances whose class is equal the key.

 If anArray contains an element that is not a kind of Behavior, an error
 is generated.

 Scans the entire Repository at least once.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 An error is raised if the argument anArray contains more than 2024 elements."


| inputSet inputSetSize inputIdx resultInSetOrder result inputArraySize
  anObj scanSetThisTime |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

inputSet := IdentitySet withAll: anArray .
inputSetSize := inputSet size .
inputSetSize < 1 ifTrue:[ ^ Array new ] .
inputSet do: [:each | each _validateClass: Behavior].

inputIdx := 1 .
resultInSetOrder := Array new .
[inputIdx <= inputSetSize] whileTrue:[
   scanSetThisTime := IdentitySet new .
   [ scanSetThisTime size < 2000 _and:[inputIdx <= inputSetSize ]] whileTrue:[
      scanSetThisTime add: (inputSet _at: inputIdx) .
      inputIdx := inputIdx + 1 .
      ].
   resultInSetOrder addAll: (self _countInstances: scanSetThisTime) .
   ].
inputArraySize := anArray size .
result := Array new: inputArraySize .
1 to: inputArraySize do:[:j|
  anObj := anArray at: j .
  result at: j put: ( resultInSetOrder at: (inputSet _offsetOf: anObj))
  ].
^ result
%

category: 'Private'
method: Repository
_countInstances: anIdentitySet

"Private.  Applications should use the Repository>>countInstances: method."

<primitive: 571>

anIdentitySet _validateClass: IdentitySet.
(anIdentitySet size > 2024 "virtual machine constant")
  ifTrue:[ anIdentitySet _error: #rtErrArgOutOfRange  ] .
anIdentitySet do: [:each | each _validateClass: Behavior].
^self _primitiveFailed: #_countInstances:
%

category: 'Private'
method: Repository
_listInstancesInPageOrder: anIdentitySet

"Private.  Applications should use the Repository>>listInstancesInPgaeOrder: method."

"Returns a list of instances on the receiver that belong to one of the classes
 listed in anArray.  The result of this method is an Array, listing all instances
 of all input classes in page-order.

 Warning: you may retrieve instances to which you have no read access, so this
 method is mostly of use to SystemUser.

 The argument aSet must not be larger than 2024 elements.

 Scans the entire Repository."

<primitive: 519>

anIdentitySet _validateClass: IdentitySet.
(anIdentitySet size > 2024 "virtual machine constant")
  ifTrue:[ anIdentitySet _error: #rtErrArgOutOfRange  ] .
anIdentitySet do: [:each | each _validateClass: Behavior].
^self _primitiveFailed: #_listInstancesInPageOrder:
%

category: 'Class Management'
method: Repository
listInstancesInPageOrder: anArray

"Returns a list of instances on the receiver that belong to one of the
 classes listed in anArray sorted by page order.  The result of this method
 is a single Array containing all instances of all input classes.

 This method is especially useful when the result will be scanned by one 
 or more sessions to perform instance migration.  Scanning the list in page
 order should perform much faster than scanning it in object ID order, which
 is how the results from listInstances: and listReferences: are returned.

                            Warning:
 You may retrieve instances to which you have no read access, so this method
 is mostly of use to SystemUser.

 If anArray contains multiple occurrences of a class, then the result
 will contain corresponding multiple occurrences of the same Set that lists
 the instances of that class.

 If anArray contains an element that is not a kind of Behavior, an error
 is generated.

 Scans the entire Repository."

| inputSet inputSetSize |

inputSet := IdentitySet withAll: anArray .
inputSetSize := inputSet size .
(inputSetSize > 2024 "virtual machine constant")
  ifTrue:[ anArray _error: #rtErrArgOutOfRange  ] .

^self _listInstancesInPageOrder: inputSet .
%

category: 'Debugging Support'
method: Repository
_analyzeDataPage: aPageId
"GemStone internal use only.  Writes output to stdout.
 Fails if not executed as SystemUser"

<primitive: 573>
aPageId _validateClass: Integer .
^ self _primitiveFailed: #_analyzeDataPage:
%

category: 'Safe Full Backups'
method: Repository
safeFullBackupCompressedTo: fileOrDevice

"Backup the receiver to a single backup file or tape in gzip format.  See
 safeFullBackupCompressedTo:MBytes: for further documentation.

 Backup files written to a file system in compressed mode will have 
 the suffix '.gz' appended to the file name if the specified file name does
 not end with that suffix."

^ self safeFullBackupTo: fileOrDevice MBytes: 0 compressed: true
%
category: 'Safe Full Backups'
method: Repository
safeFullBackupCompressedTo: fileOrDevice MBytes: mByteLimit

"Same as safeFullBackupTo:MBytes: , except that the output file is 
 written compressed in gzip format, and cannot be written to a raw device.
 The output file must be on a local file system; a pgsvr will not be used.

 Backup files written to a file system in compressed mode will have 
 the suffix '.gz' appended to the file name if the specified file name does
 not end with that suffix."

  ^ self safeFullBackupTo: fileOrDevice MBytes: mByteLimit compressed: true
%
category: 'Safe Full Backups'
method: Repository
safeFullBackupTo: fileOrDevice

"Backup the receiver to a single backup file or tape.  See
 safeFullBackupTo:MBytes: for further documentation.

 GemStone Systems has QA tested backup/restore of backup files
 up to 16GB.  For larger repositories, we recommend you use
 safeFullBackupTo:MBytes: to break the backup into several files.

 This method requires the FileControl privilege.

 A GciHardBreak during this method will terminate the session."

^ self safeFullBackupTo: fileOrDevice MBytes: 0
%
category: 'Safe Full Backups'
method: Repository
safeFullBackupTo: fileOrDevice MBytes: mByteLimit

"Produces a full backup file containing the most recently committed version of
 the receiver as of the time the method is executed.

 This method is the same as the fullBackupTo: fileOrDevice MBytes: mByteLimit
 method except that this one checks to ensure each object is valid before adding
 it to the backup file.  The 'normal' full backup only performs this check if
 the page containing the object is on the shadowed pages list.

 The argument fileOrDevice (a kind of String) specifies the file or device
 where the backup is to be created.  

 fileOrDevice can be either a file on some file system, or a device name 
 specifying a tape device.  Backups to raw partitions are no longer 
 supported.

 The fileOrDevice argument may use GemStone Network Resource String syntax.  
 For example, this may be used to access a tape device on another machine, 
 provided a GemStone NetLDI process is running on the remote machine.  

 If fileOrDevice specifies a file that already exists on a fileSystem, an
 error is generated.

 The mByteLimit argument, which specifies the maximum size of fileOrDevice in
 units of megabytes, must be a SmallInteger.  The value 0 means that there is
 no limit on the size of the resulting fileOrDevice.  A mByteLimit less than 0
 or greater than 4096000 will generate an error.  GemStone Systems has QA tested
 backup/restore of backup files up to 16GB.  While larger backup files will
 probably work, they have not been QA tested, so we recommend you limit backup
 file size to 16GB, or conduct your own tests with larger sizes before trusting
 the backup.

 If the backup requires more bytes than you specified in mByteLimit, this method
 returns a message (a String) stating that a partial backup file was created.
 In this case, further commits in this session are disallowed until
 you either complete the full backup with continueFullBackupTo:MBytes: or
 cancel the backup with abortFullBackup.  To continue the backup, you can
 execute the method continueFullBackupTo:MBytes:, which creates the next file
 in the backup sequence.

 If fileOrDevice runs out of space, such as off the end of a tape, the backup
 will terminate with a system I/O error at that point.  The backup will be
 unusable.  To avoid having to repeat the entire backup, make sure the device
 has sufficient space or set mByteLimit appropriately.

 When the size of your GemStone repository exceeds the capacity of a backup
 tape or file system, you can use mByteLimit (a SmallInteger) to control the 
 maximum number of bytes to be written to the backup file.  

 This method puts the session into auto-begin transaction mode and aborts
 the current transaction.  If an abort would cause unsaved  changes to be
 lost, it signals an error, #rtErrAbortWouldLoseData.   If the repository
 is not in restore state, then this method commits a record of the start
 of the backup to UserGlobals at: #BackupLog;  this commit is done as a
 checkpoint.   Then the transaction mode is changed to manual begin, 
 and the remainder of the backup operation executes outside of a
 transaction so that it does not cause excessive repository growth.  
 A varying number of aborts are done while outside of a transaction, 
 depending on the time required to execute the backup.

 When the backup completes, the session is always left outside of a transaction
 so that it does not retain a commit record that would cause the repository to
 grow.

 Returns true if the backup was completed.  Returns a message (a String) if
 continueFullBackupTo:MBytes: should be run to complete the backup.

 This method requires the FileControl privilege.

 A GciHardBreak during this method will terminate the session."

  ^ self safeFullBackupTo: fileOrDevice MBytes: mByteLimit compressed: false
%
category: 'Private'
method: Repository
safeFullBackupTo: fileOrDevice MBytes: mByteLimit compressed: comprBool

| result |

 System needsCommit
   ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

(SystemRepository _restoreStatusInfo at: 2) == nil ifTrue:[
   | logObj uGlobals |

   System inTransaction 
     ifFalse:[System beginTransaction ] .  "abort/begin the current transaction"
  "Repository is not in restore state, so record the start of 
   the backup in the log object"
  uGlobals := (GsSession currentSession resolveSymbol: #UserGlobals) value .
  logObj := uGlobals at:#BackupLog ifAbsent:[
    uGlobals at:#BackupLog put: String new 
    ] .
  logObj add: 'safeFullBackup to ' + fileOrDevice + ' started at ';
       add: (DateTime now asStringUsingFormat: #[2,1,3,$/,2,1,$:,true,false,false]);
       add: Character lf .
  "with tranlog/replay of dead, commit here, and abort to checkpoint in C."
  System commitTransaction 
    ifFalse:[ self _error: #rtErrMaxCommitRetry ] .
].

System inTransaction 
  ifFalse:[System beginTransaction ] .  "primitive must be called within a transaction"

result := self _fullBackupTo: fileOrDevice
               backupId: 1 
               MBytes: mByteLimit 
               firstFile: true 
               compressed: comprBool .

^result
%

category: 'Safe Garbage Collection'
method: Repository
safeMarkGcCandidates

"Obsolete method.  markGcCandidates is now always safe."

^ self markGcCandidates
%

category: 'Safe Garbage Collection'
method: Repository
safeFindDisconnectedObjects

"Same as the findDisconnectedObjects method, except that every object scanned
 is verified to be valid by checking the object table."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self _safeFindDisconnectedObjects
%

category: 'Safe Garbage Collection'
method: Repository
_safeFindDisconnectedObjects

<primitive: 575>

self _primitiveFailed: #_safeFindDisconnectedObjs.
self _uncontinuableError
%

category: 'Safe Garbage Collection'
method: Repository
safeMarkForCollection

"Same as the markForCollection method except that every object scanned is
 verified to be valid by checking the object table."

^ self safeMarkForCollectionWait: 120 

%

category: 'Safe Garbage Collection'
method: Repository
safeMarkForCollectionWait: waitTimeSeconds

"Same as the markForCollectionWait: method except that every object scanned is
 verified to be valid by checking the object table."

self prepareForMarkForCollection .
^self _primSafeMarkForCollection: waitTimeSeconds
%

category: 'Safe Garbage Collection'
method: Repository
_primSafeMarkForCollection: waitTimeSeconds

<primitive: 576>

self _primitiveFailed: #_primSafeMarkForCollection: .
self _uncontinuableError
%

category: 'Private'
method: Repository
primCleanupDeadSymbols: readOnlyBoolean

<primitive: 578>
^self _primitiveFailed: #cleanupDeadSymbols
%

category: 'Safe Garbage Collection'
method: Repository
cleanupDeadSymbols: readOnlyBoolean

"Scan the repository to locate a list of unneeded (dead) Symbols; i.e., those which
 are referenced only by the AllSymbols collection.  The list is stored in
 hidden set 26 when the method returns.

 If aBoolean is true, then the operation is read-only.  The list of dead
 symbols is created but the repository is not modified, the method
 does not commit, and AllSymbols is not locked.

 If readOnlyBoolean is false, then the method also removes the dead symbols from the
 AllSymbols collection and commits one or more transactions.  AllSymbols is also 
 write-locked for the duration of this method.  Therefore no symbols may be created
 or removed while this method is in progress.  The dead symbol objects will then be
 deleted from the repository after the garbage collection cycle.  

 Symbols with reserved oops are always considered to be live objects and are not
 returned by this method.

 Non-canonical symbols (symbols which are not in the AllSymbols collection) are 
 also detected and reported by this method.  Non-canonical symbols are never
 repaired by this method and must be repaired manually.

 Note that this method will create non-canonical symbols when run in non-read-only
 mode.  Such symbols will be removed by the next global garbage collection cycle.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.  This method
 runs inside a transaction for its entire duration and is intended to be run in
 a non-production environment.

 This method must be run as SystemUser.

 Returns an Array of 5 Integers as follows:

 1 - Number of live symbols found.
 2 - Number of dead symbols found (stored in hidden set 26).
 3 - Number of non-canonical symbols found (stored in hidden set 27).
 4 - Number of repairable non-canonical symbols found.
 5 - Number of non-repairable, non-canonical symbols found (stored in hidden set 28)."

 System needsCommit
   ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self primCleanupDeadSymbols: readOnlyBoolean
%

category: 'Private'
method: Repository
_findOopsOnPages: anArrayOfPageIds
"Scan the object table and collect all OOPs that reference any page
 in the input array of page IDs."
<primitive: 581>
self _primitiveFailed: #_findOopsOnPages: .
%

category: 'Multi-threaded Garbage Collection'
method: Repository
markForCollectionWaitMt: waitTimeSeconds pageBufferSize: numPages threads: numThreads

"This method performs the same function as the #markForCollectionWait: 
 method except it uses multiple threads.  The algorithm uses 2 page
 buffers, each of which is allocated to hold numPages pages which are read
 into the buffer by numThreads threads.  For example, if numPages is 512
 and numThreads is 8, 2 x 512 page buffers will be created and 2 x 8
 read threads will be created.

 This method will respond to SigAbort requests from the stone in a less
 timely manner than the #markForCollection method, and therefore
 may cause significant commit record backlogs.  However the stone will not
 send a LostOtRoot signal to a session while it is running this method.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 This method is not supported for remote gem sessions or gem sessions
 running non-UNIX platforms.

 Requires the GarbageCollection privilege."

self prepareForMarkForCollection .
^ self _primMarkForCollectionMt: waitTimeSeconds pageBufferSize: numPages threads: numThreads
%

category: 'Multi-threaded Garbage Collection'
method: Repository
_primMarkForCollectionMt: waitTimeSeconds pageBufferSize: numPages threads: numThreads

<primitive: 582>

waitTimeSeconds _validateClass: SmallInteger .
numPages _validateClass: SmallInteger .
numThreads _validateClass: SmallInteger .
self _primitiveFailed: #_primMarkForCollectionMt:pageBufferSize:threads: .
self _uncontinuableError
%

category: 'Multi-threaded Garbage Collection'
method: Repository
_primFindDisconnectedObjsMt: numPages threads: numThreads

<primitive: 583>

numPages _validateClass: SmallInteger .
numThreads _validateClass: SmallInteger .
self _primitiveFailed: #_primFindDisconnectedObjsMt:threads: .
self _uncontinuableError
%

category: 'Multi-threaded Garbage Collection'
method: Repository
findDisconnectedObjsMtWithPageBufferSize: numPages threads: numThreads

"This method performs the same function as the #findDisconnectedObjects
 method except it uses multiple threads.  The algorithm uses 2 page
 buffers, each of which is allocated to hold numPages pages which are read
 into the buffer by numThreads threads.  For example, if numPages is 512
 and numThreads is 8, 2 x 512 page buffers will be created and 2 x 8
 read threads will be created.

 This method will respond to SigAbort requests from the stone in a less
 timely manner than the #findDisconnectedObjects method, and therefore
 may cause significant commit record backlogs.  However the stone will not
 send a LostOtRoot signal to a session while it is running this method.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.        

 This method is not supported for remote gem sessions or gem sessions
 running non-UNIX platforms.

 Requires the GarbageCollection privilege."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _primFindDisconnectedObjsMt: numPages threads: numThreads
%

category: 'Multi-threaded Garbage Collection'
method: Repository
markForCollectionMt

"Performs a multi-threaded markForCollection with the default settings
 (640 total pages buffered, 4 read threads, and a GC lock wait time of
 120 seconds).

 Each of these settings may be specified by using the method:
    #markForCollectionWaitMt:pageBufferSize:threads:
 instead of this one."

^ self markForCollectionWaitMt: 120 pageBufferSize: 320 threads: 2
%

category: 'Multi-threaded Garbage Collection'
method: Repository
markForCollectionWaitMt: waitTimeSeconds

"Performs a multi-threaded markForCollection with the default settings
 (640 total pages buffered and 4 read threads).

 Each of these settings may be specified by using the method:
    #markForCollectionWaitMt:pageBufferSize:threads:
 instead of this one."

^ self markForCollectionWaitMt: waitTimeSeconds pageBufferSize: 320 threads: 2
%

category: 'Multi-threaded Garbage Collection'
method: Repository
supportsMtGc

"Answer true if the multi-threaded garbage collection may be run by
 this session, otherwise answers false."

System sessionIsLocal 
  ifFalse:[ ^ false ] .

System sessionIsUNIX
  ifFalse:[ ^ false ] .

^ true
%
